"""
Functionality to create contract-specific target positions from signals.
"""

import numpy as np
import pandas as pd
from typing import List, Union
from macrosynergy.management.utils import reduce_df
from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.types import QuantamentalDataFrame
from macrosynergy.panel.historic_vol import historic_vol
from macrosynergy.panel.make_zn_scores import make_zn_scores
from macrosynergy.panel.basket import Basket
from macrosynergy.compat import PD_FUTURE_STACK
import random


def target_positions(
    df: pd.DataFrame,
    cids: List[str],
    xcat_sig: str,
    ctypes: Union[List[str], str],
    sigrels: List[float],
    basket_names: Union[str, List[str]] = [],
    ret: str = "XR_NSA",
    start: str = None,
    end: str = None,
    scale: str = "prop",
    min_obs: int = 252,
    thresh: float = None,
    cs_vtarg: float = None,
    lback_periods: int = 21,
    lback_meth: str = "ma",
    half_life: int = 11,
    posname: str = "POS",
):
    """
    Converts signals into contract-specific target positions.

    Parameters
    ----------
    df : ~pandas.DataFrame
        standardized DataFrame containing at least the following columns: 'cid', 'xcat',
        'real_date' and 'value'.
    cids : List[str]
        cross-sections of markets or currency areas in which positions should be taken.
    xcat_sig : str
        category that serves as signal across markets.
    ctypes : List[str]
        contract types that are traded across markets. They should correspond to return
        categories in the dataframe if the `ret` argument is appended. Examples are 'FX' or
        'EQ'.
    sigrels : List[float]
        values that translate the single signal into contract type and basket signals in
        the order defined by keys.
    basket_names : str or List[str]
        single string or list of the names of several baskets. The weight dataframes
        will be appended to the main dataframe. Therefore, use the basket name to isolate
        the corresponding weights. The default value for the parameter is an empty list,
        which mean that no baskets are traded.
    ret : str
        postfix denoting the returns in % associated with contract types. For JPMaQS
        derivatives return data this is typically "XR_NSA" (default). Returns are required
        for volatility targeting.
    start : str
        earliest date in ISO format. Default is None and earliest date for which the
        signal category is available is used.
    end : str
        latest date in ISO format. Default is None and latest date for which the signal
        category is available is used.
    scale : str
        method that translates signals into unit target positions. The following options
        are available:

        - [1] "prop":
            means proportionate. In this case zn-scoring is applied to the signal based
            on the panel, with the neutral level set at zero. A 1 SD value translates
            into a USD1 position in the contract. This translation may apply winsorization
            through the `thresh` argument.

        - [2] "dig":
            means 'digital' and sets the individual position to either USD1 long or short,
            depending on the sign of the signal. Note that a signal of zero translates
            into a position of zero. Note that unit target positions may subsequently
            be calibrated to meet cross- section volatility targets using the `cs_targ`
            argument.

    min_obs : int
        the minimum number of observations required to calculate zn_scores. Default is
        252. Note: For the initial minimum period of the signal time series in-sample zn-
        scoring is used.
    thresh : float
        threshold value beyond which zn-scores for proportionate position taking are
        winsorized. The threshold is the maximum absolute score value in standard
        deviations. The minimum is 1 standard deviation. Default is None and means no
        winsorization.
    cs_vtarg : float
        Value for volatility targeting at the cross-section level. The purpose of this
        operation is to relate signal to risk rather than notional. Default is None and
        means no volatility targeting. If a value is chosen then for each cross-section a
        unit position is defined as a position for which the annual return standard
        deviation is equal to that value. For example, a target of 10 and a cross-section
        signal of 0.5 standard deviations would translate into a target position that
        carries a recent historical annualized standard deviation of 5 dollars (or other
        currency units).
    lback_periods : int
        Number of lookback periods over which volatility is calculated. Default is 21.
        Typically this refers to days.
    lback_meth : str
        Lookback method to calculate the volatility. Default is "ma", which means simple
        moving average. Alternative is "ema", which means exponential moving average.
    half_life : int
        Refers to the half-time for "xma". Default is 11.
    posname : str
        postfix added to contract to denote position name.

    Returns
    -------
    ~pandas.Dataframe
        standardized dataframe with daily target positions in USD, using the columns
        'cid', 'xcat', 'real_date' and 'value'.  Note: A target position differs from a
        signal insofar as it is a dollar amount and determines to what extent the size of
        signal (as opposed to direction) matters. Further, if the modified signal has a NaN
        value, the target position will be converted to zero: a position will not be taken
        given the signal was not available for that respective date. A target position also
        differs from an actual position in two ways. First, the actual position can only be
        aligned with the target with some lag. Second, the actual position will be affected
        by other considerations, such as risk management and assets under management.
    """

    # A. Initial checks
    df = QuantamentalDataFrame(df)
    _as_categorical = df.InitializedAsCategorical
    if basket_names:
        df_c_wgts, baskets = weight_dataframes(df=df, basket_names=basket_names)
        df_c_wgts = iter(df_c_wgts)
        no_panels = len(sigrels) - len(basket_names)
        panel_sigrels = sigrels[:no_panels]  # sigrels only for regular panels.
    else:
        panel_sigrels = sigrels

    if isinstance(ctypes, str):
        ctypes = [ctypes]
    ctypes_baskets = ctypes + basket_names

    cols = ["cid", "xcat", "real_date", "value"]
    assert set(cols) <= set(df.columns), f"df columns must contain {cols}."

    categories = set(df["xcat"].unique())
    error_1 = "Signal category missing from the standardised dataframe."
    assert xcat_sig in categories, error_1
    error_2 = "Volatility Target must be numeric value."
    if cs_vtarg is not None:
        assert isinstance(cs_vtarg, (float, int)), error_2

    error_3 = (
        "The number of signal relations must be equal to the number of contracts "
        "and, if defined, the number of baskets defined in 'ctypes'."
    )
    clause = len(ctypes_baskets)
    assert len(sigrels) == clause, error_3
    assert isinstance(min_obs, int), "Minimum observation parameter must be an integer."

    # B. Reduce frame to necessary data.

    df = df.loc[:, cols]
    contract_returns = [c + ret for c in ctypes]
    xcats = contract_returns + [xcat_sig]

    dfx = reduce_df(df=df, xcats=xcats, cids=cids, start=start, end=end, blacklist=None)

    # C. Calculate and reformat modified cross-sectional signals.

    df_mods = modify_signals(
        df=dfx,
        cids=cids,
        xcat_sig=xcat_sig,
        start=start,
        end=end,
        scale=scale,
        min_obs=min_obs,
        thresh=thresh,
    )  # (USD 1 per SD or sign).

    df_mods_w = df_mods.pivot(index="real_date", columns="cid", values="value")

    # D. Volatility target ratios (if required).

    use_vtr = False
    if isinstance(cs_vtarg, (int, float)):
        # D.1. Composite signal-related positions as basis for volatility targeting.

        df_csurs = cs_unit_returns(
            dfx, contract_returns=contract_returns, sigrels=panel_sigrels
        )  # Gives cross-section returns.

        # D.2. Calculate volatility adjustment ratios.

        df_vol = historic_vol(
            df_csurs,
            xcat=ret,
            cids=cids,
            lback_periods=lback_periods,
            lback_meth=lback_meth,
            half_life=half_life,
            start=start,
            end=end,
            remove_zeros=True,
            postfix="",
        )  # Gives unit position vols.

        dfw_vol = df_vol.pivot(index="real_date", columns="cid", values="value")
        dfw_vol = dfw_vol.sort_index(axis=1)
        dfw_vtr = 100 * cs_vtarg / dfw_vol  # vol-target ratio to be applied.
        use_vtr = True

    # E. Actual position calculation.

    df_pos_cons = []
    ctypes_sigrels = dict(zip(ctypes_baskets, sigrels))

    for k, v in ctypes_sigrels.items():
        df_mods_copy = df_mods_w.copy()

        if use_vtr:
            dfw_pos_vt = df_mods_copy.multiply(dfw_vtr)
            dfw_pos_vt.dropna(how="all", inplace=True)
            df_mods_copy = dfw_pos_vt

        if k in basket_names:
            contracts = baskets[k]
            df_c_weights = next(df_c_wgts)
            df_mods_copy = basket_handler(
                df_mods_w=df_mods_copy, df_c_wgts=df_c_weights, contracts=contracts
            )

        # Allows for the signal being applied to the basket constituents on the original
        # dataframe.
        df_mods_copy *= v  # modified signal x sigrel = post-VT position.

        df_posi = df_mods_copy.stack(**PD_FUTURE_STACK).to_frame("value").reset_index()
        df_posi["value"] = df_posi["value"].fillna(0)
        df_posi = QuantamentalDataFrame.from_long_df(df_posi, xcat=k)
        df_pos_cons.append(df_posi)

    # Consolidate the positions across the formed panels and baskets (baskets will be a
    # subset of the panels).
    if basket_names:
        df_pos_cons = consolidate_positions(df_pos_cons, ctypes)
    df_tpos = pd.concat(df_pos_cons, axis=0, ignore_index=True)

    df_tpos = QuantamentalDataFrame(df_tpos).rename_xcats(postfix=f"_{posname}")
    df_tpos = reduce_df(df=df_tpos, xcats=None, cids=None, start=start, end=end)

    df_tpos = df_tpos.reset_index(drop=True)
    df_tpos = QuantamentalDataFrame(
        df_tpos, _initialized_as_categorical=_as_categorical
    ).to_original_dtypes()
    return df_tpos


def weight_dataframes(df: pd.DataFrame, basket_names: Union[str, List[str]] = None):
    """
    Helper function used to break up the original dataframe and create the wide weight
    dataframes.

    Parameters
    ----------
    df : ~pandas.DataFrame
        standardized DataFrame.
    basket_names : str or List[str]
        string or list of basket names.

    Returns
    -------
    List[pd.DataFrames], dict
    """

    if isinstance(basket_names, str):
        basket_names = [basket_names]

    xcats = df["xcat"].to_numpy()
    wgt_indices = lambda index: index.split("_")[-1] == "WGT"
    boolean = list(map(wgt_indices, xcats))

    r_df = df[boolean].copy()
    b_column = lambda index: "_".join(index.split("_")[1:-1])

    r_df.loc[:, "basket_name"] = np.array(list(map(b_column, r_df["xcat"])))

    df_c_wgts = []
    b_dict = {}
    contr_func = lambda index: "_".join(index.split("_")[0:2])

    r_df_copy = r_df.copy()

    for b in basket_names:
        b_df = r_df[r_df["basket_name"] == b]
        b_df_copy = r_df_copy[r_df_copy["basket_name"] == b]
        b_df_copy = b_df_copy.drop(["basket_name"], axis=1)
        column = b_df["cid"] + "_" + b_df["xcat"]
        b_df_copy["xcat"] = np.array(list(map(contr_func, column)))
        b_wgt = b_df_copy.pivot(index="real_date", columns="xcat", values="value")
        b_wgt = b_wgt.reindex(sorted(b_wgt.columns), axis=1)
        df_c_wgts.append(b_wgt)
        contracts = set(b_df_copy["xcat"])
        b_dict[b] = sorted(list(contracts))

    return df_c_wgts, b_dict


def modify_signals(
    df: pd.DataFrame,
    cids: List[str],
    xcat_sig: str,
    start: str = None,
    end: str = None,
    scale: str = "prop",
    min_obs: int = 252,
    thresh: float = None,
    blacklist: Union[dict, None] = None
):
    """
    Calculate modified cross-section signals based on zn-scoring (proportionate method)
    or conversion to signs (digital method).

    Parameters
    ----------
    df : ~pandas.DataFrame
        standardized DataFrame containing the following columns: 'cid', 'xcat',
        'real_date' and 'value'.
    cids : List[str]
        cross sections of markets or currency areas in which positions should be taken.
    xcat_sig : str
        category that serves as signal across markets.
    start : str
        earliest date in ISO format. Default is None and earliest date for which the
        signal category is available is used.
    end : str
        latest date in ISO format. Default is None and latest date for which the signal
        category is available is used.
    scale : str
        method that translates signals into unit target positions. The following options
        are available:

        - [1] "prop":
            means proportionate. In this case zn-scoring is applied to the signal based
            on the panel, with the neutral level set at zero. A 1 SD value translates
            into a USD1 position in the contract. This translation may apply winsorization
            through the `thresh` argument.

        - [2] "dig":
            means 'digital' and sets the individual position to either USD1 long or short,
            depending on the sign of the signal. Note that a signal of zero translates
            into a position of zero. Note that unit target positions may subsequently
            be calibrated to meet cross- section volatility targets using the `cs_targ`
            argument.

    min_obs : int
        the minimum number of observations required to calculate zn_scores. Default is
        252. Note: For the initial period of the signal time series in-sample zn-scoring is
        used.
    thresh : float
        threshold value beyond which zn-scores for propotionate position taking are
        winsorized. The threshold is the maximum absolute score value in standard
        deviations. The minimum is 1 standard deviation.
    blacklist : dict or None
        dictionary of cross-sections and their respective date ranges that should be
        excluded from the calculation of the modified signals. The keys are cross-section 
        identifiers (e.g., 'AUD') and the values are lists of date ranges (e.g., ['2012-01-03', '2012-01-03']).

    Returns
    -------
    ~pandas.Dataframe
        standardized dataframe, of modified signaks, using the columns 'cid', 'xcat',
        'real_date' and 'value'.
    """

    options = ["prop", "dig"]
    assert scale in options, f"The scale parameter must be either {options}"

    if scale == "prop":
        df_ms = make_zn_scores(
            df,
            xcat=xcat_sig,
            sequential=True,
            cids=cids,
            start=start,
            end=end,
            neutral="zero",
            pan_weight=1,
            min_obs=min_obs,
            iis=True,
            thresh=thresh,
            blacklist=blacklist,
        )
    else:
        df_ms = reduce_df(
            df=df, xcats=[xcat_sig], cids=cids, start=start, end=end, blacklist=blacklist
        )
        df_ms["value"] = np.sign(df_ms["value"])

    return df_ms


def cs_unit_returns(
    df: pd.DataFrame,
    contract_returns: List[str],
    sigrels: List[str],
    ret: str = "XR_NSA",
):
    """
    Calculate returns of composite unit positions (that jointly depend on one signal).

    Parameters
    ----------
    df : ~pandas.DataFrame
        standardized DataFrame containing the following columns: 'cid', 'xcat',
        'real_date' and 'value'.
    contract_returns : List[str]
        list of the contract return types.
    sigrels : List[float]
        respective signal for each contract type.
    ret : str
        postfix denoting the returns in % applied to the contract types. Default is
        "XR_NSA".

    Returns
    -------
    ~pandas.Dataframe
        standardized dataframe with the summed portfolio returns which are used to
        calculate the evolving volatility, using the columns 'cid', 'xcat', 'real_date' and
        'value'.
    """

    error_message = "Each individual contract requires an associated signal."
    assert len(contract_returns) == len(sigrels), error_message
    df_c_rets = None

    for i, c_ret in enumerate(contract_returns):
        # Filter the DataFrame 'df' to just c_ret xcats
        df_c_ret = df[df["xcat"] == c_ret]
        # Reshape the DataFrame by setting 'real_date' as index, 'cid' as columns, and
        # 'value' as values.
        df_c_ret = df_c_ret.pivot(index="real_date", columns="cid", values="value")

        # Sort the dataframe and multiply by the respective signal.
        df_c_ret = df_c_ret.sort_index(axis=1)
        df_c_ret *= sigrels[i]

        df_c_rets = df_c_ret if i == 0 else df_c_rets + df_c_ret

    # Any dates not shared by all categories will be removed.
    df_c_rets.dropna(how="all", inplace=True)

    df_rets = df_c_rets.stack().to_frame("value").reset_index()
    df_rets = QuantamentalDataFrame.from_long_df(df_rets, xcat=ret)

    cols = ["cid", "xcat", "real_date", "value"]

    return df_rets[cols].sort_values(by=cols[:3])


def basket_handler(
    df_mods_w: pd.DataFrame, df_c_wgts: pd.DataFrame, contracts: List[str]
):
    """
    Function designed to compute the target positions for the constituents of a basket.
    The function will return the corresponding basket dataframe of positions.

    Parameters
    ----------
    df_mods_w : ~pandas.DataFrame
        target position dataframe. Will be multiplied by the weight dataframe to
        establish the positions for the basket of constituents.
    df_c_wgts : ~pandas.DataFrame
        weight dataframe used to adjust the positions of the basket of contracts.
    contracts : dict
        the constituents that make up each basket.

    Returns
    -------
    ~pandas.Dataframe
        basket positions weight-adjusted.
    """

    error_1 = "df_c_wgts expects to receive a pd.DataFrame."
    assert isinstance(df_c_wgts, pd.DataFrame), error_1
    error_2 = (
        "df_c_wgts expects a pivoted pd.DataFrame - each column corresponds to the"
        " contract's weight."
    )
    assert df_c_wgts.index.name == "real_date", error_2
    error_3 = (
        f"df_c_wgts column names must correspond to the received contract: "
        f"{contracts}."
    )
    assert all(df_c_wgts.columns == contracts), error_3

    split = lambda b: b.split("_")[0]

    cross_sections = list(map(split, contracts))

    # Sort the columns to align via cross-sections to conduct the multiplication. The
    # weight dataframe is formed using the respective contracts, so additional checks are
    # not required.
    dfw_wgs = df_c_wgts.reindex(sorted(df_c_wgts.columns), axis=1)

    # Reduce to the cross-sections held in the respective basket.
    df_mods_w = df_mods_w[cross_sections]
    df_mods_w = df_mods_w.reindex(sorted(df_mods_w.columns), axis=1)

    # Adjust the target positions to reflect the weighting method. Align the pandas names
    # to allow for pd.DataFrame.multiply().
    dfw_wgs.columns = df_mods_w.columns
    df_mods_w = df_mods_w.multiply(dfw_wgs)

    return df_mods_w


def date_alignment(panel_df: pd.DataFrame, basket_df: pd.DataFrame):
    """
    Method used to align the panel position dataframe and the basket dataframe of
    weight-adjusted positions to the same timeframe.

    Parameters
    ----------
    panel_df : ~pandas.DataFrame

    basket_df : ~pandas.DataFrame


    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame]
        returns the two received dataframes defined over the same period.
    """

    p_dates = panel_df["real_date"].to_numpy()
    b_dates = basket_df["real_date"].to_numpy()
    if p_dates[0] > b_dates[0]:
        index = np.searchsorted(b_dates, p_dates[0])
        basket_df = basket_df.iloc[index:, :]
    elif p_dates[0] < b_dates[0]:
        index = np.searchsorted(p_dates, b_dates[0])
        panel_df = panel_df.iloc[index:, :]
    else:
        pass

    if p_dates[-1] > b_dates[-1]:
        index = np.searchsorted(p_dates, b_dates[-1], side="right")
        panel_df = panel_df.iloc[:index, :]
    elif p_dates[-1] < b_dates[-1]:
        index = np.searchsorted(b_dates, p_dates[-1], side="right")
        basket_df = basket_df.iloc[:index, :]
    else:
        pass

    return panel_df, basket_df


def consolidation_help(panel_df: pd.DataFrame, basket_df: pd.DataFrame):
    """
    The function receives a panel dataframe and a basket of cross-sections of the same
    contract type. Therefore, aim to consolidate the targeted positions across the
    shared contracts.

    Parameters
    ----------
    panel_df : ~pandas.DataFrame
        The dataframe of the panel positions.
    basket_df : ~pandas.DataFrame
        The dataframe of the basket positions.

    Returns
    -------
    Tuple[pd.DataFrame, pd.DataFrame]
        returns the consolidated and reduced dataframes.
    """

    basket_cids = basket_df["cid"].unique()
    panel_cids = panel_df["cid"].unique()

    panel_copy = []
    for cid in panel_cids:
        indices = panel_df["cid"] == cid
        temp_df = panel_df[indices]

        if cid in basket_cids:
            basket_indices = basket_df["cid"] == cid
            basket_rows = basket_df[basket_indices]
            temp_df, basket_rows = date_alignment(temp_df, basket_rows)
            b_values = basket_rows["value"].to_numpy()

            panel_values = temp_df["value"].to_numpy()
            consolidation = panel_values + b_values
            temp_df["value"] = consolidation
            panel_copy.append(temp_df)

            basket_indices = ~basket_indices
            basket_df = basket_df[basket_indices]
        else:
            panel_copy.append(temp_df)

    return pd.concat(panel_copy)


def consolidate_positions(data_frames: List[pd.DataFrame], ctypes: List[str]):
    """
    Method used to consolidate positions if baskets are used. The constituents of a
    basket will be a subset of one of the panels.

    Parameters
    ----------
    data_frames : List[~pandas.DataFrame]
        list of the target position dataframes.
    ctypes : List[str]
        list of the contract types.

    Returns
    -------
    List[~pandas.DataFrame]
        list of dataframes having consolidated positions.
    """

    no_ctypes = len(ctypes)
    dict_ = dict(zip(ctypes[:no_ctypes], data_frames[:no_ctypes]))
    df_baskets = data_frames[no_ctypes:]
    # Iterating exclusively through the basket dataframes.
    for df in df_baskets:
        category = df["xcat"].str.split("_", expand=True)[1]
        c_type = category.iloc[0]

        panel_df = dict_[c_type]
        dict_[c_type] = consolidation_help(panel_df, basket_df=df)

    return list(dict_.values())


if __name__ == "__main__":
    # A. Example dataframe

    cids = ["AUD", "GBP", "NZD", "USD"]
    xcats = ["FXXR_NSA", "EQXR_NSA", "SIG_NSA"]

    ccols = ["earliest", "latest", "mean_add", "sd_mult"]
    df_cids = pd.DataFrame(index=cids, columns=ccols)
    df_cids.loc["AUD"] = ["2010-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2010-01-01", "2020-12-31", 0, 2]
    df_cids.loc["NZD"] = ["2010-01-01", "2020-12-31", 0, 3]
    df_cids.loc["USD"] = ["2010-01-01", "2020-12-31", 0, 4]

    xcols = ccols + ["ar_coef", "back_coef"]
    df_xcats = pd.DataFrame(index=xcats, columns=xcols)
    df_xcats.loc["FXXR_NSA"] = ["2010-01-01", "2020-12-31", 0, 1, 0, 0.2]
    df_xcats.loc["EQXR_NSA"] = ["2010-01-01", "2020-12-31", 0.5, 2, 0, 0.2]
    df_xcats.loc["SIG_NSA"] = ["2010-01-01", "2020-12-31", 0, 10, 0.4, 0.2]

    random.seed(2)
    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd_copy = dfd.copy()
    black = {"AUD": ["2000-01-01", "2003-12-31"], "GBP": ["2018-01-01", "2100-01-01"]}

    # B. Target positions without basket

    df1 = target_positions(
        df=dfd,
        cids=cids,
        xcat_sig="SIG_NSA",
        ctypes=["FX", "EQ"],
        sigrels=[1, 0.5],
        ret="XR_NSA",
        start="2012-01-01",
        end="2020-10-30",
        scale="prop",
        min_obs=252,
        cs_vtarg=5,
        posname="POS",
    )

    df2 = target_positions(
        df=dfd,
        cids=cids,
        xcat_sig="FXXR_NSA",
        ctypes=["FX", "EQ"],
        sigrels=[1, -1],
        ret="XR_NSA",
        start="2012-01-01",
        end="2020-10-30",
        scale="dig",
        cs_vtarg=0.1,
        posname="POS",
    )

    df3 = target_positions(
        df=dfd,
        cids=cids,
        xcat_sig="FXXR_NSA",
        ctypes=["FX", "EQ"],
        sigrels=[1, -1],
        ret="XR_NSA",
        start="2010-01-01",
        end="2020-12-31",
        scale="prop",
        cs_vtarg=None,
        posname="POS",
    )

    # C. Target position with one basket

    apc_contracts = ["AUD_FX", "NZD_FX"]
    basket_1 = Basket(
        df=dfd, contracts=apc_contracts, ret="XR_NSA", cry=None, blacklist=black
    )
    basket_1.make_basket(weight_meth="equal", max_weight=0.55, basket_name="APC_FX")
    df_weight = basket_1.return_weights("APC_FX")

    df_weight = df_weight[["cid", "xcat", "real_date", "value"]]
    dfd = dfd[["cid", "xcat", "real_date", "value"]]
    dfd_concat = pd.concat([dfd_copy, df_weight])

    df4 = target_positions(
        df=dfd_concat,
        cids=cids,
        xcat_sig="SIG_NSA",
        ctypes=["FX", "EQ"],
        basket_names=["APC_FX"],
        sigrels=[1, -1, -0.5],
        ret="XR_NSA",
        start="2010-01-01",
        end="2020-12-31",
        scale="prop",
        cs_vtarg=10,
        posname="POS",
    )



from macrosynergy.learning.preprocessing.scalers.base_panel_scaler import (
    BasePanelScaler,
)


class PanelMinMaxScaler(BasePanelScaler):
    """
    Scale and translate panel features to lie within the range [0,1].

    Notes
    -----
    This class is designed to replicate scikit-learn's `MinMaxScaler()` class, with the
    additional option to scale within cross-sections. Unlike the `MinMaxScaler()` class,
    dataframes are always returned, preserving the multi-indexing of the inputs.
    """

    def extract_statistics(self, X, feature):
        """
        Determine the minimum and maximum values of a feature in the input matrix.

        Parameters
        ----------
        X : pandas.DataFrame
            The feature matrix.
        feature : str
            The feature to extract statistics for.

        Returns
        -------
        statistics : list
            List containing the minimum and maximum values of the feature.
        """
        return [X[feature].min(), X[feature].max()]

    def scale(self, X, feature, statistics):
        """
        Scale the 'feature' column in the design matrix 'X' based on the minimum and
        maximum values of the feature.

        Parameters
        ----------
        X : pandas.DataFrame
            The feature matrix.
        feature : str
            The feature to scale.
        statistics : list
            List containing the minimum and maximum values of the feature, in that order.

        Returns
        -------
        X_transformed : pandas.Series
            The scaled feature.
        """
        return (X[feature] - statistics[0]) / (statistics[1] - statistics[0])


class PanelStandardScaler(BasePanelScaler):
    """
    Scale and translate panel features to have zero mean and unit variance.

    Parameters
    ----------
    type : str, default="panel"
        The panel dimension over which the scaling is applied. Options are
        "panel" and "cross_section".
    with_mean : bool, default=True
        Whether to centre the data before scaling.
    with_std : bool, default=True
        Whether to scale the data to unit variance.

    Notes
    -----
    This class is designed to replicate scikit-learn's StandardScaler() class, with the
    additional option to scale within cross-sections. Unlike the StandardScaler() class,
    dataframes are always returned, preserving the multi-indexing of the inputs.
    """

    def __init__(self, type="panel", with_mean=True, with_std=True):
        # Checks
        if not isinstance(with_mean, bool):
            raise TypeError("'with_mean' must be a boolean.")
        if not isinstance(with_std, bool):
            raise TypeError("'with_std' must be a boolean.")

        # Attributes
        self.with_mean = with_mean
        self.with_std = with_std

        super().__init__(type=type)

    def extract_statistics(self, X, feature):
        """
        Determine the mean and standard deviation of values of a feature in the input
        matrix.

        Parameters
        ----------
        X : pandas.DataFrame
            The feature matrix.
        feature : str
            The feature to extract statistics for.

        Returns
        -------
        statistics : list
            List containing the mean and standard deviation of values of the feature.
        """
        return [X[feature].mean(), X[feature].std()]

    def scale(self, X, feature, statistics):
        """
        Scale the 'feature' column in the design matrix 'X' based on the mean and
        standard deviation values of the feature.

        Parameters
        ----------
        X : pandas.DataFrame
            The feature matrix.
        feature : str
            The feature to scale.
        statistics : list
            List containing the mean and standard deviation of values of the feature,
            in that order.

        Returns
        -------
        X_transformed : pandas.Series
            The scaled feature.
        """
        if self.with_mean:
            if self.with_std:
                return (X[feature] - statistics[0]) / statistics[1]
            else:
                return X[feature] - statistics[0]
        else:
            if self.with_std:
                return X[feature] / statistics[1]
            else:
                return X[feature]


if __name__ == "__main__":
    import numpy as np
    import pandas as pd

    from macrosynergy.management import (
        categories_df,
        make_qdf,
    )

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    """Example: Unbalanced panel """

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2002-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2003-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2000-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2000-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2000-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2000-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2000-01-01", "2020-12-31", -0.1, 2, 0.8, 0.3]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {
        "GBP": (
            pd.Timestamp(year=2009, month=1, day=1),
            pd.Timestamp(year=2012, month=6, day=30),
        ),
        "CAD": (
            pd.Timestamp(year=2015, month=1, day=1),
            pd.Timestamp(year=2100, month=1, day=1),
        ),
    }

    train = categories_df(
        df=dfd, xcats=xcats, cids=cids, val="value", blacklist=black, freq="M", lag=1
    ).dropna()
    train = train[
        train.index.get_level_values(1) >= pd.Timestamp(year=2005, month=8, day=1)
    ]

    X_train = train.drop(columns=["XR"])
    y_train = train["XR"]

    # Standard scaling over each cross-section
    scaler = PanelStandardScaler(type="cross_section")
    scaler.fit(X_train, y_train)
    print(scaler.transform(X_train))

    # MinMax scaling over the panel
    scaler = PanelMinMaxScaler()
    scaler.fit(X_train, y_train)
    print(scaler.transform(X_train))



from .base_panel_selector import BasePanelSelector
from .panel_selectors import (
    LarsSelector,
    LassoSelector,
    MapSelector,
)
__all__ = [
    "BasePanelSelector",
    "LarsSelector",
    "LassoSelector",
    "MapSelector",
]


"""
JPMaQS Download Interface
"""

import datetime
from dateutil.relativedelta import relativedelta
import io
import logging
import os
import glob
import shutil
import json
import warnings
from timeit import default_timer as timer
from typing import Dict, List, Optional, Tuple, Union, Any, Generator
import itertools
import joblib
from tqdm import tqdm

import pandas as pd

from macrosynergy.download.dataquery import (
    JPMAQS_GROUP_ID,
    DataQueryInterface,
    API_DELAY_PARAM,
)
from macrosynergy.download.exceptions import InvalidDataframeError, DataOutOfSyncError
from macrosynergy.management.utils import (
    is_valid_iso_date,
    concat_single_metric_qdfs,
    ticker_df_to_qdf,
)
from macrosynergy.management.constants import JPMAQS_METRICS
from macrosynergy.management.types import QuantamentalDataFrame

logger = logging.getLogger(__name__)
debug_stream_handler = logging.StreamHandler(io.StringIO())
debug_stream_handler.setLevel(logging.NOTSET)
debug_stream_handler.setFormatter(
    logging.Formatter(
        "%(asctime)s - %(levelname)s - %(module)s - %(funcName)s :: %(message)s"
    )
)
logger.addHandler(debug_stream_handler)

DEFAULT_CLIENT_ID_ENV_VAR: str = "DQ_CLIENT_ID"
DEFAULT_CLIENT_SECRET_ENV_VAR: str = "DQ_CLIENT_SECRET"


def deconstruct_expression(
    expression: Union[str, List[str]],
) -> Union[List[str], List[List[str]]]:
    """
    Deconstruct an expression into a list of cid, xcat, and metric. Achieves the inverse
    of construct_expressions(). For non-JPMaQS expressions, the returned list will be
    [expression, expression, 'value']. The metric is set to 'value' to ensure the
    reported metric is consistent with the standard JPMaQS metrics
    (JPMaQSDownload.valid_metrics).

    Parameters
    ----------
    expression : str
        expression to deconstruct. If a list is provided, each element will be
        deconstructed and returned as a list of lists.

    Raises
    ------
    TypeError
        if `expression` is not a string or a list of strings.
    ValueError
        if `expression` is an empty list.

    Returns
    -------
    list[str]
        list of cid, xcat, and metric.
    """

    if not isinstance(expression, (str, list)):
        raise TypeError("`expression` must be a string or a list of strings.")

    if isinstance(expression, list):
        if not all(isinstance(exprx, str) for exprx in expression):
            raise TypeError("All elements of `expression` must be strings.")
        elif len(expression) == 0:
            raise ValueError("`expression` must be a non-empty list.")
        return list(map(deconstruct_expression, expression))
    else:
        try:
            exprx: str = expression.replace("DB(JPMAQS,", "").replace(")", "")
            ticker, metric = exprx.split(",")
            result: List[str] = ticker.split("_", 1) + [metric]
            if len(result) != 3:
                raise ValueError(f"{exprx} is not a valid JPMaQS expression.")
            return ticker.split("_", 1) + [metric]
        except Exception as e:
            warnings.warn(
                f"Failed to deconstruct expression `{expression}`: assuming it is a non-JPMaQS expression.",
                UserWarning,
            )
            # fail safely, return list where cid = xcat = expression,
            #  and metric = 'value'
            return [expression, expression, "value"]


def check_attributes_in_sync(ts_list) -> bool:
    """
    Checks if the attributes in the response are in sync with the time-series data. This
    is performed since on occasion the ticker will have just been calculated for a new
    date but on certain pods the data won't have updated yet but on some it will have
    updated. This can lead to the attributes on a specific time-series being out of
    sync.

    Parameters
    ----------
    response_dict : list
        List containing the response from the API.

    Returns
    -------
    bool
        True if the attributes are in sync, False otherwise.
    """

    expressions_last_value_dict = {}

    for instrument in ts_list:
        attributes = instrument.get("attributes")
        if not attributes:
            continue

        time_series = attributes[0].get("time-series")
        if not time_series:
            continue

        last_valid_item = None
        for i in range(len(time_series) - 1, -1, -1):
            if time_series[i][1] is not None:
                last_valid_item = time_series[i]
                break

        if not last_valid_item:
            last_valid_item = time_series[0]

        expression = attributes[0].get("expression")
        if not "JPMAQS" in expression:
            continue
        if not expression:
            last_valid_item = ["No data", 0]
        else:
            cid, xcat, _ = deconstruct_expression(expression)
            ticker = cid + "_" + xcat

        last_value_date = last_valid_item[0]
        if ticker not in expressions_last_value_dict:
            expressions_last_value_dict[ticker] = last_value_date
        else:
            if last_value_date != expressions_last_value_dict[ticker]:
                return False

    return True


def construct_expressions(
    tickers: Optional[List[str]] = None,
    cids: Optional[List[str]] = None,
    xcats: Optional[List[str]] = None,
    metrics: Optional[List[str]] = None,
) -> List[str]:
    """
    Construct expressions from the provided arguments.

    Parameters
    ----------
    tickers : list[str]
        list of tickers.
    cids : list[str]
        list of cids.
    xcats : list[str]
        list of xcats.
    metrics : list[str]
        list of metrics.

    Returns
    -------
    list[str]
        list of expressions.
    """

    if tickers is None:
        tickers = []
    if cids is not None and xcats is not None:
        tickers += [f"{cid}_{xcat}" for cid in cids for xcat in xcats]

    return [f"DB(JPMAQS,{tick},{metric})" for tick in tickers for metric in metrics]


def get_expression_from_qdf(df: Union[pd.DataFrame, List[pd.DataFrame]]) -> List[str]:
    if isinstance(df, list):
        return list(itertools.chain.from_iterable(map(get_expression_from_qdf, df)))

    metrics = list(set(df.columns) - set(QuantamentalDataFrame.IndexCols))
    exprs = []
    for metric in metrics:
        for cid, xcat in df[["cid", "xcat"]].drop_duplicates().values:
            if any(~df.loc[(df["cid"] == cid) & (df["xcat"] == xcat), metric].isna()):
                exprs.append(f"DB(JPMAQS,{cid}_{xcat},{metric})")
    return exprs


def get_expression_from_wide_df(
    df: Union[pd.DataFrame, List[pd.DataFrame]],
) -> List[str]:
    if isinstance(df, list):
        return list(itertools.chain.from_iterable(map(get_expression_from_wide_df, df)))
    return list(set(df.columns))


def timeseries_to_qdf(timeseries: Dict[str, Any]) -> QuantamentalDataFrame:
    """
    Converts a dictionary containing a time-series to a QuantamentalDataFrame.

    Parameters
    ----------
    timeseries : Dict[str, Any]
        A dictionary containing a time-series.

    Returns
    -------
    QuantamentalDataFrame
        The converted DataFrame.
    """

    if not isinstance(timeseries, dict):
        raise TypeError("Argument `timeseries` must be a dictionary.")

    if _get_ts(timeseries) is None:
        return None

    cid, xcat, metric = deconstruct_expression(_get_expr(timeseries))
    df = pd.DataFrame(_get_ts(timeseries), columns=["real_date", metric])
    df["real_date"] = pd.to_datetime(df["real_date"], format="%Y%m%d")

    if df.empty or all(df.isna().all()):
        return None

    df = QuantamentalDataFrame.from_long_df(
        df=df.dropna().reset_index(drop=True),
        value_column=metric,
        cid=cid,
        xcat=xcat,
    )

    if metric == "last_updated":
        df["last_updated"] = pd.to_datetime(df["last_updated"])

    return df


def timeseries_to_column(
    timeseries: Dict[str, Any], errors: str = "ignore"
) -> pd.DataFrame:
    """
    Converts a dictionary of time series to a DataFrame with a single column.

    Parameters
    ----------
    timeseries : Dict[str, Any]
        A dictionary of time series.
    errors : str
        The error handling method to use. If 'raise', then invalid items in the list
        will raise an error. If 'ignore', then invalid items will be ignored. Default is
        'ignore'.

    Returns
    -------
    pd.DataFrame
        The converted DataFrame.
    """

    if not isinstance(timeseries, dict):
        raise TypeError("Argument `timeseries` must be a dictionary.")

    if errors not in ["raise", "ignore"]:
        raise ValueError("`errors` must be one of 'raise' or 'ignore'.")

    expression = _get_expr(timeseries)

    if _get_ts(timeseries) is None:
        if errors == "raise":
            raise ValueError("Time series is empty.")
        return None

    df: pd.DataFrame = pd.DataFrame(
        _get_ts(timeseries), columns=["real_date", expression]
    )
    df["real_date"] = pd.to_datetime(df["real_date"], format="%Y%m%d")
    df = df.dropna()
    if df.empty:
        if errors == "raise":
            raise ValueError("Time series is empty.")
        return None
    return df.set_index("real_date")


def concat_column_dfs(
    df_list: List[pd.DataFrame], errors: str = "ignore"
) -> pd.DataFrame:
    """
    Concatenates a list of DataFrames into a single DataFrame.

    Parameters
    ----------
    df_list : List[pd.DataFrame]
        A list of DataFrames.
    errors : str
        The error handling method to use. If 'raise', then invalid items in the list
        will raise an error. If 'ignore', then invalid items will be ignored. Default is
        'ignore'.

    Returns
    -------
    pd.DataFrame
        The concatenated DataFrame.
    """

    if not isinstance(df_list, list):
        raise TypeError("Argument `df_list` must be a list.")

    if errors not in ["raise", "ignore"]:
        raise ValueError("`errors` must be one of 'raise' or 'ignore'.")

    if not all([isinstance(df, pd.DataFrame) for df in df_list]):
        if errors == "raise":
            raise TypeError("All elements in `df_list` must be DataFrames.")
        df_list = [df for df in df_list if isinstance(df, pd.DataFrame)]

    def _pop_df_list() -> Generator[pd.DataFrame, None, None]:
        while df_list:
            yield df_list.pop(0)

    # all the dfs are indexed by real_date, so we can just concat them adn drop dates when all values are NaN
    # df: pd.DataFrame = pd.concat(df_list, axis=1).dropna(how="all", axis=0)
    return pd.concat(_pop_df_list(), axis=1).dropna(how="all", axis=0)


def _get_expr(ts: dict) -> str:
    return ts["attributes"][0]["expression"]


def _get_ts(ts: dict) -> dict:
    return ts["attributes"][0]["time-series"]


def _get_ticker(ts: dict) -> str:
    return _get_expr(ts).split(",")[1]


def _get_xcat(ticker: str) -> str:
    return ticker.split("_", 1)[1]


def _ticker_filename(ticker: str, save_path: str) -> str:
    return os.path.join(save_path, _get_xcat(ticker), f"{ticker}.csv")


def _save_qdf(data: List[dict], save_path: str) -> None:
    for ticker in sorted(set(map(_get_ticker, data))):
        ticker_filename = _ticker_filename(ticker, save_path)
        os.makedirs(os.path.dirname(ticker_filename), exist_ok=True)
        ts = [_ts for _ts in data if _get_ticker(_ts) == ticker]
        df: QuantamentalDataFrame = QuantamentalDataFrame.from_qdf_list(
            [timeseries_to_qdf(_ts) for _ts in ts]
        ).drop(columns=["cid", "xcat"])

        if os.path.exists(_ticker_filename(ticker, save_path)):
            edf = pd.read_csv(
                ticker_filename, parse_dates=["real_date"], index_col="real_date"
            )
            edf = edf.drop(columns=[col for col in edf.columns if col in df.columns])
            df = pd.concat([edf, df.set_index("real_date")], axis=1).reset_index()
            os.remove(ticker_filename)
        df.to_csv(ticker_filename, index=False)

    return


def _save_timeseries_as_column(data: List[dict], save_path: str) -> None:
    for ts in data:
        if _get_ts(ts) is None:
            continue
        expr = _get_expr(ts)
        df = timeseries_to_column(ts)
        if df.empty:
            continue
        df.reset_index().to_csv(os.path.join(save_path, f"{expr}.csv"), index=False)
    return


def _save_timeseries(data: List[dict], save_path: str):
    data = [d for d in data if d is not None]
    if len(data) == 0:
        return

    for ts in data:
        with open(os.path.join(save_path, f"{_get_expr(ts)}.json"), "w") as f:
            json.dump(ts, f)

    return


def validate_downloaded_df(
    data_df: pd.DataFrame,
    expected_expressions: List[str],
    found_expressions: List[str],
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    verbose: bool = True,
) -> bool:
    """
    Validate the downloaded data in the provided dataframe.

    Parameters
    ----------
    data_df : pd.DataFrame
        dataframe containing the downloaded data.
    expected_expressions : list[str]
        list of expressions that were expected to be downloaded.
    found_expressions : list[str]
        list of expressions that were actually downloaded.
    start_date : str
        start date of the downloaded data.
    end_date : str
        end date of the downloaded data.
    verbose : bool
        whether to print the validation results.

    Raises
    ------
    TypeError
        if `data_df` is not a dataframe.

    Returns
    -------
    bool
        True if the downloaded data is valid, False otherwise.
    """

    if not isinstance(data_df, pd.DataFrame):
        raise InvalidDataframeError(
            "Empty or invalid dataframe, please check download parameters."
        )
    if data_df.empty:
        return False

    # check if all expressions are present in the df
    exprs_f, expr_expected = set(found_expressions), set(expected_expressions)
    expr_missing = expr_expected - exprs_f
    unexpected_exprs = exprs_f - expr_expected
    if unexpected_exprs:
        raise InvalidDataframeError(
            f"Unexpected expressions were found in the downloaded data: "
            f"{unexpected_exprs}"
        )

    if expr_missing:
        log_str = (
            f"Some expressions are missing from the downloaded data. "
            "Check logger output for complete list.\n"
            f"{len(expr_missing)} out of {len(expr_expected)} expressions are missing. "
            f"To download the catalogue of all available expressions and filter the "
            "unavailable expressions, set `get_catalogue=True` in the "
            "call to `JPMaQSDownload.download()`."
        )

        logger.info(log_str)
        logger.info(f"Missing expressions: {expr_missing}")
        if verbose:
            print(log_str)

    # check if all dates are present in the df
    # NOTE : Hardcoded max start date to 1990-01-01. This is because the JPMAQS
    # database does not have data before this date.
    if datetime.datetime.strptime(start_date, "%Y-%m-%d") < datetime.datetime.strptime(
        "1990-01-01", "%Y-%m-%d"
    ):
        start_date = "1950-01-01"

    dates_expected = pd.bdate_range(start=start_date, end=end_date)
    found_dates = (
        data_df["real_date"].unique()
        if isinstance(data_df, QuantamentalDataFrame)
        else data_df.index.unique()
    )

    dates_missing = list(set(dates_expected) - set(pd.to_datetime(found_dates)))
    log_str = (
        "The expressions in the downloaded data are not a subset of the expected expressions."
        " Missing expressions: {missing_exprs}"
    )
    err_statement = (
        "The expressions in the downloaded data are not a subset of the "
        "expected expressions."
    )
    check_exprs = set()
    if isinstance(data_df, QuantamentalDataFrame):
        found_metrics = list(
            set(data_df.columns) - set(QuantamentalDataFrame.IndexCols)
        )
        for col in QuantamentalDataFrame.IndexCols:
            if not len(data_df[col].unique()) > 0:
                raise InvalidDataframeError(f"Column {col} is empty.")
        tkrs = (
            (data_df["cid"].astype(str) + "_" + data_df["xcat"].astype(str))
            .unique()
            .tolist()
        )
        check_exprs = construct_expressions(tickers=tkrs, metrics=found_metrics)

    else:
        check_exprs = data_df.columns.tolist()

    missing_exprs = set(check_exprs) - set(found_expressions)
    if len(missing_exprs) > 0:
        logger.critical(log_str.format(missing_exprs=missing_exprs))

    if len(dates_missing) > 0:
        log_str = (
            f"Some dates are missing from the downloaded data. \n"
            f"{len(dates_missing)} out of {len(dates_expected)} dates are missing."
        )
        logger.warning(log_str)
        if verbose:
            print(log_str)

    return True


def _get_expressions_from_qdf_csv(file_path: str) -> List[str]:
    ticker = os.path.basename(file_path).split(".")[0]
    with open(file_path, "r", encoding="utf-8") as f:
        headers = f.readline().strip().split(",")

        assert len(set(headers)) == len(headers), f"Duplicate headers in {file_path}"
        metrics = set(headers) - set(["real_date"])
        return [f"DB(JPMAQS,{ticker},{metric})" for metric in metrics]


def _get_expressions_from_wide_csv(file_path: str) -> List[str]:
    with open(file_path, "r", encoding="utf-8") as f:
        headers = f.readline().strip().split(",")
        assert len(set(headers)) == len(headers), f"Duplicate headers in {file_path}"
        expression = list(set(headers) - set(["real_date"]))
        return expression


def _get_expressions_from_json(file_path: str) -> List[str]:
    with open(file_path, "r", encoding="utf-8") as f:
        return [_get_expr(json.load(f))]


def get_expressions_from_file(
    file_path: str, as_dataframe: bool = True, dataframe_format: str = "qdf"
) -> List[str]:
    """
    Loads the expressions found in a downloaded timeseries file (either JSON or CSV).

    Parameters
    ----------
    file_path : str
        path to the file.
    as_dataframe : bool
        whether to load the file as a dataframe.
    dataframe_format : str
        the format of the dataframe. Must be one of 'qdf' or 'wide'.

    Returns
    -------
    List[str]
        list of expressions found in the file.
    """

    if not os.path.exists(file_path):
        raise FileNotFoundError(f"File {file_path} does not exist.")

    if not as_dataframe:
        return _get_expressions_from_json(file_path)

    if not dataframe_format in ["qdf", "wide"]:
        raise ValueError("`dataframe_format` must be one of 'qdf' or 'wide'.")

    if dataframe_format == "qdf":
        return _get_expressions_from_qdf_csv(file_path)
    elif dataframe_format == "wide":
        return _get_expressions_from_wide_csv(file_path)


def validate_downloaded_data(
    path: str,
    expected_expressions: List[str],
    as_dataframe: bool = True,
    dataframe_format: str = "qdf",
    show_progress: bool = True,
) -> List[str]:
    """
    Validate the downloaded data in the provided path.

    Parameters
    ----------
    path : str
        path to the downloaded data.
    expected_expressions : list[str]
        list of expressions that were expected to be downloaded.
    as_dataframe : bool
        whether to load the files as dataframes.
    dataframe_format : str
        the format of the dataframe. Must be one of 'qdf' or 'wide'.
    show_progress : bool
        whether to show a progress bar.

    Returns
    -------
    list[str]
        list of expressions that are missing from the downloaded data.
    """

    if not os.path.isdir(path):
        raise ValueError(f"Path {path} does not exist.")

    ext = "csv" if as_dataframe else "json"
    files = glob.glob(f"{path}/**/*.{ext}", recursive=True)

    def get_expression_func(
        file_path: str, as_dataframe=as_dataframe, dataframe_format=dataframe_format
    ) -> List[str]:
        return get_expressions_from_file(
            file_path, as_dataframe=as_dataframe, dataframe_format=dataframe_format
        )

    all_exprs = []
    all_exprs = joblib.Parallel(n_jobs=-1)(
        joblib.delayed(get_expression_func)(file_path)
        for file_path in tqdm(
            files,
            desc="Validating downloaded data",
            disable=not show_progress,
        )
    )
    # join all the lists of expressions
    all_exprs = list(itertools.chain.from_iterable(all_exprs))

    missing_exprs = sorted(set(expected_expressions) - set(all_exprs))

    if len(missing_exprs) > 0:
        logger.critical(
            f"Some expressions are missing from the downloaded data. "
            f"Missing expressions: {missing_exprs}"
        )

    return missing_exprs


class JPMaQSDownload(DataQueryInterface):
    """
    JPMaQSDownload Object. This object is used to download JPMaQS data via the DataQuery
    API. It can be extended to include the use of proxies, and even request generic
    DataQuery expressions.

    Parameters
    ----------
    oauth : bool
        True if using oauth, False if using username/password with crt/key.
    client_id : Optional[str]
        oauth client_id, required if oauth=True.
    client_secret : Optional[str]
        oauth client_secret, required if oauth=True.
    crt : Optional[str]
        path to crt file.
    key : Optional[str]
        path to key file.
    username : Optional[str]
        username for certificate based authentication.
    password : Optional[str]
        paired with username for certificate
    debug : bool
        True if debug mode, False if not.
    suppress_warning : bool
        True if suppressing warnings, False if not.
    check_connection : bool
        True if the interface should check the connection to the server before sending
        requests, False if not. False by default.
    proxy : Optional[dict]
        proxy to use for requests, None if not using proxy (default).
    print_debug_data : bool
        True if debug data should be printed, False if not (default).
    dq_kwargs : dict
        additional arguments to pass to the DataQuery API object such `calender` and
        `frequency` for the DataQuery API. For more fine-grained usage, initialize the
        DataQueryInterface object explicitly.
    kwargs : dict
        any other keyword arguments.

    Raises
    ------
    TypeError
        if provided arguments are not of the correct type.
    ValueError
        if provided arguments are invalid or semantically incorrect.
    """

    def __init__(
        self,
        oauth: bool = True,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        crt: Optional[str] = None,
        key: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        check_connection: bool = True,
        proxy: Optional[Dict] = None,
        suppress_warning: bool = True,
        debug: bool = False,
        print_debug_data: bool = False,
        dq_download_kwargs: dict = {},
        *args,
        **kwargs,
    ):
        vars_types_zip: List[Tuple[str, str]] = [
            (oauth, "oauth", bool),
            (check_connection, "check_connection", bool),
            (suppress_warning, "suppress_warning", bool),
            (debug, "debug", bool),
            (print_debug_data, "print_debug_data", bool),
        ]

        for varx, namex, typex in vars_types_zip:
            if not isinstance(varx, typex):
                raise TypeError(f"`{namex}` must be of type {typex}.")

        if not isinstance(proxy, dict) and proxy is not None:
            raise TypeError("`proxy` must be a dictionary or None.")

        if not isinstance(dq_download_kwargs, dict):
            raise TypeError("`dq_download_kwargs` must be a dictionary.")

        self.suppress_warning = suppress_warning
        self.debug = debug
        self.print_debug_data = print_debug_data
        self._check_connection = check_connection
        self.dq_download_kwargs = dq_download_kwargs

        for varx, namex in [
            (client_id, "client_id"),
            (client_secret, "client_secret"),
            (crt, "crt"),
            (key, "key"),
            (username, "username"),
            (password, "password"),
        ]:
            if varx is not None:
                if not isinstance(varx, str):
                    raise TypeError(f"`{namex}` must be a string.")

        if not all([crt, key, username, password]):
            if not all([client_id, client_secret]):
                # check the environment variables
                _clid = os.getenv(DEFAULT_CLIENT_ID_ENV_VAR)
                _clsc = os.getenv(DEFAULT_CLIENT_SECRET_ENV_VAR)
                if all([_clid, _clsc]):
                    client_id = _clid
                    client_secret = _clsc

        if not (all([client_id, client_secret]) or all([crt, key, username, password])):
            raise ValueError(
                "Must provide either `client_id` and `client_secret` for oauth, or "
                "`crt`, `key`, `username`, and `password` for certificate based authentication."
            )

        self.valid_metrics: List[str] = JPMAQS_METRICS
        self.msg_errors: List[str] = []
        self.msg_warnings: List[str] = []
        self.unavailable_expressions: List[str] = []
        self.downloaded_data: Dict = {}
        self.jpmaqs_access: bool = True

        super().__init__(
            oauth=oauth,
            client_id=client_id,
            client_secret=client_secret,
            crt=crt,
            key=key,
            username=username,
            password=password,
            proxy=proxy,
            check_connection=check_connection,
            suppress_warning=suppress_warning,
            debug=debug,
            *args,
            **kwargs,
        )

        if self._check_connection:
            self.check_connection()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback): ...

    def _get_unavailable_expressions(
        self,
        expected_exprs: List[str],
        dicts_list: Optional[List[dict]] = None,
        downloaded_df: Optional[Union[pd.DataFrame, QuantamentalDataFrame]] = None,
    ) -> List[str]:
        assert (dicts_list is not None) ^ (downloaded_df is not None)
        if dicts_list is not None:
            return super()._get_unavailable_expressions(
                expected_exprs=expected_exprs, dicts_list=dicts_list
            )

        if downloaded_df is not None:
            if len(downloaded_df) == 0:
                return expected_exprs
            if isinstance(downloaded_df, QuantamentalDataFrame):
                found_expressions = get_expression_from_qdf(downloaded_df)
            else:
                found_expressions = get_expression_from_wide_df(downloaded_df)
            return list(set(expected_exprs) - set(found_expressions))

    def validate_download_args(
        self,
        tickers: List[str],
        cids: List[str],
        xcats: List[str],
        metrics: List[str],
        start_date: str,
        end_date: str,
        get_catalogue: bool,
        expressions: List[str],
        show_progress: bool,
        as_dataframe: bool,
        dataframe_format: str,
        report_time_taken: bool,
    ) -> bool:
        """
        Validate the arguments passed to the download function.

        Raises
        ------
        TypeError
            If any of the arguments are not of the correct type.
        ValueError
            If any of the arguments are semantically incorrect.

        Returns
        -------
        bool
            True if valid.
        """

        for var, name in [
            (get_catalogue, "get_catalogue"),
            (show_progress, "show_progress"),
            (as_dataframe, "as_dataframe"),
            (report_time_taken, "report_time_taken"),
        ]:
            if not isinstance(var, bool):
                raise TypeError(f"`{name}` must be a boolean.")

        if not isinstance(dataframe_format, str):
            raise TypeError("`dataframe_format` must be a string.")
        elif dataframe_format.lower() not in ["qdf", "wide"]:
            raise ValueError("`dataframe_format` must be one of 'qdf' or 'wide'.")

        if all([tickers is None, cids is None, xcats is None, expressions is None]):
            raise ValueError(
                "Must provide at least one of `tickers`, "
                "`expressions`, or `cids` & `xcats` together."
            )

        for var, name in [
            (tickers, "tickers"),
            (cids, "cids"),
            (xcats, "xcats"),
            (expressions, "expressions"),
        ]:
            if not isinstance(var, list) and var is not None:
                raise TypeError(f"`{name}` must be a list of strings.")
            if var is not None:
                if len(var) > 0:
                    if not all([isinstance(ticker, str) for ticker in var]):
                        raise TypeError(f"`{name}` must be a list of strings.")
                else:
                    raise ValueError(f"`{name}` must be a non-empty list of strings.")

        if metrics is None:
            raise ValueError("`metrics` must be a non-empty list of strings.")
        else:
            if all([metric not in self.valid_metrics for metric in metrics]):
                raise ValueError(f"`metrics` must be a subset of {self.valid_metrics}.")

        if bool(cids) ^ bool(xcats):
            raise ValueError(
                "If specifying `xcats`, `cids` must also be specified (and vice versa)."
            )

        for var, name in [
            (start_date, "start_date"),
            (end_date, "end_date"),
        ]:
            if not is_valid_iso_date(var):  # type check covered by `is_valid_iso_date`
                raise ValueError(
                    f"`{name}` must be a valid date in the format YYYY-MM-DD."
                )
            if pd.to_datetime(var, errors="coerce") is pd.NaT:
                raise ValueError(
                    f"`{name}` must be a valid date > "
                    f"{pd.Timestamp.min.strftime('%Y-%m-%d')}.\n"
                    "Check pandas documentation:"
                    " https://pandas.pydata.org/docs/user_guide/timeseries.html#timestamp-limitations`"
                )
            if pd.to_datetime(var) < pd.to_datetime("1950-01-01"):
                warnings.warn(
                    message=(
                        f"`{name}` is set before 1950-01-01."
                        "Data before 1950-01-01 may not be available,"
                        " and will cause errors/missing data."
                    ),
                    category=UserWarning,
                )

        return True

    def filter_expressions_from_catalogue(
        self, expressions: List[str], verbose: bool = True
    ) -> List[str]:
        """
        Method to filter a list of expressions against the JPMaQS catalogue. This avoids
        requesting data for expressions that are not in the catalogue, and provides the
        user wuth the complete list of expressions that are in the catalogue.

        Parameters
        ----------
        tickers : List[str]
            list of tickers to filter.

        Returns
        -------
        List[str]
            list of tickers that are in the JPMaQS catalogue.
        """

        catalogue_tickers: List[str] = self.get_catalogue(verbose=verbose)
        catalogue_expressions: List[str] = construct_expressions(
            tickers=catalogue_tickers, metrics=self.valid_metrics
        )
        upper_exprs = [ex.upper() for ex in catalogue_expressions]
        r: List[str] = sorted(
            set(ex for ex in expressions if ex.upper() in upper_exprs)
        )
        if verbose:
            filtered: int = len(expressions) - len(r)
            if filtered > 0:
                print(
                    f"Removed {filtered}/{len(expressions)} expressions "
                    "that are not in the JPMaQS catalogue."
                )

        return r

    def _chain_download_outputs(
        self, download_outputs: Union[List[Dict], List[pd.DataFrame]]
    ) -> Union[List[Dict], List[pd.DataFrame]]:
        if not isinstance(download_outputs, list):
            raise TypeError("`download_outputs` must be a list.")

        download_outputs = [x for x in download_outputs if len(x) > 0]
        if len(download_outputs) == 0:
            return []

        if isinstance(download_outputs[0], pd.DataFrame):
            return concat_column_dfs(df_list=download_outputs)

        if isinstance(download_outputs[0][0], (dict, bool, QuantamentalDataFrame)):
            logger.debug(f"Chaining {len(download_outputs)} outputs.")
            _ch_types = list(
                itertools.chain.from_iterable(
                    [list(map(type, x)) for x in download_outputs]
                )
            )
            logger.debug(f"Object types in the downloaded data: {_ch_types}")

            download_outputs = list(itertools.chain.from_iterable(download_outputs))

        if isinstance(download_outputs[0], (dict, bool)):
            return download_outputs
        if isinstance(download_outputs[0], QuantamentalDataFrame):
            # return concat_single_metric_qdfs(download_outputs)
            return QuantamentalDataFrame.from_qdf_list(download_outputs)
            # cannot chain QDFs with different metrics
        if not self.jpmaqs_access:
            raise ValueError(
                f"The credentials you have provided are for Dataquery access only and "
                f"hence have no JPMaQS entitlements. Because of this you are only able "
                f"to download data from 2000-01-01 to "
                f"{(datetime.datetime.now() - relativedelta(months=6)).strftime('%Y-%m-%d')}."
            )

        raise NotImplementedError(
            f"Cannot chain download outputs that are List of : {list(set(map(type, download_outputs)))}."
        )

    def _save_data(
        self,
        data: List[dict],
        as_dataframe: bool,
        dataframe_format: str,
        save_path: str,
    ) -> bool:
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        if as_dataframe:
            if dataframe_format == "qdf":
                _save_qdf(data, save_path)
            elif dataframe_format == "wide":
                _save_timeseries_as_column(data, save_path)
        else:
            _save_timeseries(data, save_path)
        return True

    def _fetch_timeseries(
        self,
        url: str,
        params: dict,
        tracking_id: str,
        as_dataframe: bool = True,
        dataframe_format: str = "qdf",
        save_path: Optional[str] = None,
        *args,
        **kwargs,
    ) -> Union[pd.DataFrame, List[Dict]]:
        ts_list: List[dict] = self._fetch(
            url=url, params=params, tracking_id=tracking_id
        )
        if not check_attributes_in_sync(ts_list):
            expressions = [ts["attributes"][0]["expression"] for ts in ts_list]
            error_str = f"Attributes for {expressions} are not in sync."
            raise DataOutOfSyncError(error_str)

        for its, ts in enumerate(ts_list):
            if _get_ts(ts) is None:
                self.unavailable_expressions.append(_get_expr(ts))
                if "message" in ts["attributes"][0]:
                    self.msg_warnings.append(ts["attributes"][0]["message"])
                else:
                    self.msg_warnings.append(
                        f"Time series for expression {ts['attributes'][0]['expression']} is empty. "
                        " No explanation was provided."
                    )
                ts_list[its] = None

            if "message" in ts["attributes"][0]:
                if "[FT] Limited Dataset Access Only" in ts["attributes"][0]["message"]:
                    self.jpmaqs_access = False

        ts_list: List[dict] = list(filter(None, ts_list))
        logger.debug(f"Downloaded data for {len(ts_list)} expressions.")
        logger.debug(f"Unavailable expressions: {self.unavailable_expressions}")
        if save_path is not None:
            try:
                ts_list = [
                    self._save_data(
                        data=ts_list,
                        as_dataframe=as_dataframe,
                        dataframe_format=dataframe_format,
                        save_path=save_path,
                        *args,
                        **kwargs,
                    )
                ]
            except Exception as exc:
                logger.error(f"Failed to save data to disk: {exc}")
                self.msg_errors.append(f"Failed to save data to disk: {exc}")
                raise exc
        elif as_dataframe:
            if dataframe_format == "qdf":
                ts_list = [timeseries_to_qdf(ts) for ts in ts_list if ts is not None]
            elif dataframe_format == "wide":
                ts_list = concat_column_dfs(
                    df_list=[timeseries_to_column(ts) for ts in ts_list]
                )

        downloaded_types = list(set(map(type, ts_list)))
        logger.debug(
            f"Object types in the downloaded data: {downloaded_types}"
            + ("(saving to disk)" if save_path is not None else "")
        )

        return ts_list

    def get_catalogue(
        self,
        group_id: str = JPMAQS_GROUP_ID,
        page_size: int = 1000,
        verbose: bool = True,
    ) -> List[str]:
        """
        Get the JPMaQS catalogue.

        Returns
        -------
        List[str]
            list of tickers in the JPMaQS catalogue.
        """
        return super().get_catalogue(group_id, page_size, verbose)

    def download_all_to_disk(
        self,
        path: str,
        expressions: Optional[List[str]] = None,
        as_dataframe: bool = True,
        dataframe_format: str = "qdf",
        show_progress: bool = True,
        delay_param: float = API_DELAY_PARAM,
        batch_size: Optional[int] = None,
        retry: int = 3,
        overwrite: bool = True,
        *args,
        **kwargs,
    ) -> None:
        """
        Downloads all JPMaQS data to disk.

        Parameters
        ----------
        path : str
            path to the directory where the data will be saved.
        expressions : Optional[List[str]
            Default is None, meaning all expressions in the JPMaQS catalogue will be
            downloaded. If provided, only the expressions in the list will be downloaded.
        as_dataframe : bool
            Default is True, meaning the data will be saved as a DataFrame (either in
            the Quantamental Data Format ('qdf') or wide format ('wide')). If False, the
            data will be saved as JSON files, with one expression per file.
        dataframe_format : str
            Default is 'qdf'. If `as_dataframe` is True, this parameter specifies the
            format of the DataFrame. Must be one of 'qdf' or 'wide'.
        show_progress : bool
            Default is True, meaning the progress of the download will be displayed. If
            False, the progress will not be displayed.
        delay_param : float
            Default is 0.2 seconds (fastest allowed by DataQuery API). The delay
            parameter to use when making requests to the DataQuery API. Ideally, this should
            not be changed.
        batch_size : int
            Default is None, meaning the batch size will be set to the default size
            (20). If provided, this parameter specifies the number of expressions to
            download in each batch.
        retry : int
            Default is 3, meaning the download will be retried 3 times for any
            expressions that fail to download. If set to 0, no retries will be attempted.
        overwrite : bool
            Default is True, meaning the data will be overwritten if it already exists.
            If False, the data will not be overwritten.
        kwargs : dict
            any other keyword arguments.

        Returns
        -------
        None
            The data is saved to disk.


        Examples
        --------

        Download all JPMaQS data to disk.

        >>> with JPMaQSDownload(
        ...     client_id=os.getenv("DQ_CLIENT_ID"),
        ...     client_secret=os.getenv("DQ_CLIENT_SECRET"),
        ... ) as jpmaqs:
        ...     jpmaqs.download_all_to_disk(path="./jpmaqs-data")


        Alternatively downloading only a custom list of expressions

        >>> expressions = ['DB(JPMAQS,USD_EQXR_NSA,value)', 'DB(JPMAQS,GBP_EQXR_NSA,value)']
        >>> with JPMaQSDownload(
        ...     client_id=os.getenv("DQ_CLIENT_ID"),
        ...     client_secret=os.getenv("DQ_CLIENT_SECRET"),
        ... ) as jpmaqs:
        ...     jpmaqs.download_all_to_disk(path="./jpmaqs-data", expressions=expressions)


        Save each expression as a JSON

        >>> with JPMaQSDownload(
        ...     client_id=os.getenv("DQ_CLIENT_ID"),
        ...     client_secret=os.getenv("DQ_CLIENT_SECRET"),
        ... ) as jpmaqs:
        ...     jpmaqs.download_all_to_disk(path="./jpmaqs-data", as_dataframe=False)


        """

        save_path: Optional[str] = None
        if path == "":
            print(
                "Path explicitly provided as an empty string. "
                "Assuming alternate saving method implemented."
            )
            save_path = ""
        else:
            path = os.path.expandvars(os.path.expanduser(path))
            save_path = os.path.join(path, "JPMaQSData")
            os.makedirs(save_path, exist_ok=True)
            if overwrite:
                msg = f"Overwriting data in {save_path}."
                warnings.warn(msg)  # the user should be warned
                logger.info(msg)  # but log doesn't need to be warning, info is fine
                shutil.rmtree(save_path)
                os.makedirs(save_path, exist_ok=True)

            print(f"Downloading all JPMaQS data to disk. Saving to: `{save_path}`.")

        start_date = "1990-01-01"
        end_date = (datetime.datetime.today() + pd.offsets.BusinessDay(2)).strftime(
            "%Y-%m-%d"
        )
        self.batch_size = batch_size or self.batch_size

        self.check_connection(verbose=True)

        if not expressions:
            catalogue: List[str] = self.get_catalogue()
            expressions = sorted(
                construct_expressions(tickers=catalogue, metrics=self.valid_metrics)
            )

        # if all the expressions do not contain DB(JPMAQS, then we need set dataframe format to wide
        if (
            as_dataframe
            and dataframe_format == "qdf"
            and not all([expr.startswith("DB(JPMAQS") for expr in expressions])
        ):
            dataframe_format = "wide"
            warnings.warn(
                "The list of expressions contains non-JPMAQS expressions. "
                "Setting dataframe format to 'wide'."
            )
        print(
            "Downloading data from JPMaQS.\nTimestamp UTC: ",
            datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
        )
        download_time_taken: float = timer()
        data: List[Union[dict, QuantamentalDataFrame]] = self.download_data(
            save_path=save_path,
            expressions=expressions,
            start_date=start_date,
            end_date=end_date,
            show_progress=show_progress,
            as_dataframe=as_dataframe,
            dataframe_format=dataframe_format,
            delay_param=delay_param,
            *args,
            **kwargs,
        )
        download_time_taken: float = timer() - download_time_taken
        assert all([d is True for d in data])
        if path == "":
            return
        d_exprs = [
            os.path.basename(csv).split(".")[0]
            for csv in glob.glob(f"{save_path}/**/*.csv", recursive=True)
        ]
        if len(d_exprs) == 0:
            raise ValueError("No data was downloaded.")

        if as_dataframe and dataframe_format == "qdf":
            d_exprs = construct_expressions(tickers=d_exprs, metrics=self.valid_metrics)

        logger.info(f"Downloaded {len(d_exprs)} expressions.")

        unavailable_expressions = validate_downloaded_data(
            path=save_path,
            expected_expressions=expressions,
            as_dataframe=as_dataframe,
            dataframe_format=dataframe_format,
            show_progress=show_progress,
        )

        if len(unavailable_expressions) > 0:
            if retry > 0:
                logger.info(
                    f"Retrying {len(unavailable_expressions)} unavailable expressions."
                )
                self.download_all_to_disk(
                    path=path,
                    expressions=unavailable_expressions,
                    as_dataframe=as_dataframe,
                    dataframe_format=dataframe_format,
                    show_progress=show_progress,
                    delay_param=delay_param,
                    batch_size=batch_size,
                    retry=retry - 1,
                    overwrite=False,
                    *args,
                    **kwargs,
                )
            else:
                print(f"Failed to download {len(unavailable_expressions)} expressions.")
                for expr in unavailable_expressions:
                    print(f"\t{expr}")

    def download(
        self,
        tickers: Optional[List[str]] = None,
        cids: Optional[List[str]] = None,
        xcats: Optional[List[str]] = None,
        metrics: List[str] = ["value"],
        start_date: str = "2000-01-01",
        end_date: Optional[str] = None,
        expressions: Optional[List[str]] = None,
        get_catalogue: bool = False,
        show_progress: bool = False,
        debug: bool = False,
        suppress_warning: bool = False,
        as_dataframe: bool = True,
        dataframe_format: str = "qdf",
        report_time_taken: bool = False,
        categorical_dataframe: bool = False,
        *args,
        **kwargs,
    ) -> Union[pd.DataFrame, List[Dict]]:
        """
        Driver function to download data from JPMaQS via the DataQuery API. Timeseries
        data can be requested using `tickers` with `metrics`, or passing formed
        DataQuery expressions. `cids` and `xcats` (along with `metrics`) are used to
        construct expressions, which are ultimately passed to the DataQuery Interface.

        Parameters
        ----------
        tickers : list[str]
            list of tickers.
        cids : list[str]
            list of cids.
        xcats : list[str]
            list of xcats.
        metrics : list[str]
            list of metrics. Available metrics are "value" (default), "grading",
            "eop_lag", "mop_lag", and "last_updated". If "all" is provided,
            all available metrics are used. The available metrics are defined in
            `macrosynergy.download.jpmaqs.JPMAQS_METRICS`.
        start_date : str
            start date of the data to download, in the ISO format - YYYY-MM-DD.
        end_date : str
            end date of the data to download in the ISO format - YYYY-MM-DD.
        expressions : list[str]
            list of DataQuery expressions.
        get_catalogue : bool
            If True, the JPMaQS catalogue is downloaded and used to filter the list of
            tickers. Default is False.
        show_progress : bool
            True if progress bar should be shown, False if not (default).
        suppress_warning : bool
            True if suppressing warnings. Default is True.
        debug : bool
            Override the debug behaviour of the JPMaQSDownload class. If True, debug
            mode is enabled.
        print_debug_data : bool
            True if debug data should be printed, False if not (default). If debug=True,
            this is set to True.
        as_dataframe : bool
            Return a dataframe if True (default), a list of dictionaries if False.
        dataframe_format : str
            Format of the dataframe to return, one of "qdf" or "wide". QDF is the
            Quantamental Dataframe format, and wide is the wide format with each expression
            as a column, and a single date column.
        report_time_taken : bool
            If True, the time taken to download and apply data transformations is
            reported.
        categorical_dataframe : bool
            If True, the dataframe returned will use the pandas Categorical data type
            for the `cid` and `xcat` columns. Default is False.
        kwargs : dict
            any other keyword arguments.

        Raises
        ------
        ValueError
            if provided arguments are invalid or semantically incorrect (see
            macrosynergy.download.jpmaqs.JPMaQSDownload.validate_download_args()).

        Returns
        -------
        pd.DataFrame|list[Dict]
            dataframe of data if `as_dataframe` is True, list of dictionaries if False.
        """

        # override the default warning behaviour and debug behaviour
        self.suppress_warning = suppress_warning
        self.debug = debug

        vartolist = lambda x: [x] if isinstance(x, str) else x
        tickers = vartolist(tickers)
        cids = vartolist(cids)
        xcats = vartolist(xcats)
        expressions = vartolist(expressions)
        metrics = vartolist(metrics)

        if len(metrics) == 1:
            if metrics[0] == "all":
                metrics = self.valid_metrics

        if end_date is None:
            _today = datetime.datetime.today()
            end_date = pd.bdate_range(end=_today, periods=2)[0].strftime("%Y-%m-%d")
            # This is date is cast to YYYYMMDD in macrosynergy.download.dataquery.py.

        # Validate arguments.
        if not self.validate_download_args(
            tickers=tickers,
            cids=cids,
            xcats=xcats,
            metrics=metrics,
            start_date=start_date,
            end_date=end_date,
            expressions=expressions,
            get_catalogue=get_catalogue,
            show_progress=show_progress,
            as_dataframe=as_dataframe,
            dataframe_format=dataframe_format,
            report_time_taken=report_time_taken,
        ):
            raise ValueError("Invalid arguments passed to download().")
        if pd.to_datetime(start_date) > pd.to_datetime(end_date):
            warnings.warn(
                message=(
                    f"`start_date` ({start_date}) is after `end_date` ({end_date}). "
                    "These dates will be swapped."
                ),
                category=UserWarning,
            )
            start_date, end_date = end_date, start_date

        dataframe_format = dataframe_format.lower()
        # Construct expressions.
        if expressions is None:
            expressions: List[str] = []

        expressions += construct_expressions(
            tickers=tickers,
            cids=cids,
            xcats=xcats,
            metrics=metrics,
        )
        expressions = list(set(expressions))

        if get_catalogue:
            expressions = self.filter_expressions_from_catalogue(expressions)

        # Download data.
        print(
            "Downloading data from JPMaQS.\nTimestamp UTC: ",
            datetime.datetime.now(datetime.timezone.utc).strftime("%Y-%m-%d %H:%M:%S"),
        )
        data: List[Dict] = []
        download_time_taken: float = timer()
        data: List[Union[dict, QuantamentalDataFrame]] = self.download_data(
            expressions=expressions,
            start_date=start_date,
            end_date=end_date,
            show_progress=show_progress,
            as_dataframe=as_dataframe,
            dataframe_format=dataframe_format,
            *args,
            **kwargs,
        )

        if not self.jpmaqs_access:
            print(
                "Credentials provided only have access to Dataquery and have not been "
                "granted access to JPMaQS. You can only access data after 2000-01-01 and "
                "before 6 months from the current date."
            )

        download_time_taken: float = timer() - download_time_taken

        if report_time_taken:
            print(f"Time taken to download data: \t{download_time_taken:.2f} seconds.")

        if len(self.msg_errors) > 0:
            if not self.suppress_warning:
                warnings.warn(
                    f"{len(self.msg_errors)} errors encountered during the download. \n"
                    f"The errors did not compromise the download. \n"
                    f"Please check `JPMaQSDownload.msg_errors` for more information."
                )

        if as_dataframe:
            found_expressions: List[str] = list(
                set(expressions) - set(self.unavailable_expressions)
            )

            if not validate_downloaded_df(
                data_df=data,
                expected_expressions=expressions,
                found_expressions=found_expressions,
                start_date=start_date,
                end_date=end_date,
                verbose=True,
            ):
                raise InvalidDataframeError("Downloaded data is invalid.")

            if dataframe_format == "qdf":
                assert isinstance(data, QuantamentalDataFrame)

                return pd.DataFrame(
                    QuantamentalDataFrame(data, categorical=categorical_dataframe)
                )

        return data


def custom_download(
    tickers, download_func, metrics=["value"], start_date=None, end_date=None
):
    """
    Custom download function to download data for a list of tickers using a custom
    download function.

    Parameters
    ----------
    tickers : list[str]
        list of tickers to download data for.
    download_func : callable
        custom download function.
    metrics : list[str]
        list of metrics to download.
    start_date : str
        start date of the data to download.
    end_date : str
        end date of the data to download.

    Returns
    -------
    pd.DataFrame
        dataframe of downloaded data.
    """

    dfs = []
    for metric in metrics:
        expressions = []
        for ticker in list(set(tickers)):
            dq_expr = f"DB(JPMAQS,{ticker},{metric})"
            expressions.append(dq_expr)
        df = pd.DataFrame()

        step_size = 100
        df_store = []
        for idx in range(0, len(expressions) + 1, step_size):
            df_chunk = download_func(
                expressions[idx : idx + step_size],
                startDate=start_date,
                endDate=end_date,
            )
            df_chunk = df_chunk.dropna(axis=1, how="all")
            df_store.append(df_chunk)
        df = pd.concat(df_store, axis=1)

        df.columns = df.columns.str.split(",").str[1]
        df.index.name = "real_date"

        df = ticker_df_to_qdf(df, metric=metric)

        dfs.append(df)

    df = concat_single_metric_qdfs(dfs)
    return df


if __name__ == "__main__":
    cids = ["AUD", "BRL", "CAD", "CHF", "CNY", "CZK", "EUR", "GBP", "USD"]
    xcats = ["RIR_NSA", "FXXR_NSA", "FXXR_VT10", "DU05YXR_NSA", "DU05YXR_VT10"]
    start_date: str = "2024-01-07"
    end_date: str = "2024-02-09"

    with JPMaQSDownload(
        client_id=os.getenv("DQ_CLIENT_ID"),
        client_secret=os.getenv("DQ_CLIENT_SECRET")
    ) as jpmaqs:
        data = jpmaqs.download(
            xcats=xcats,
            cids=cids,
            metrics="all",
            start_date=start_date,
            end_date=end_date,
            show_progress=True,
            report_time_taken=True,
        )
        print(data.info())
        print(data)



"""
Functions used to visualize autocorrelation and partial autocorrelation functions.
"""

from typing import Callable, Dict, List, Optional, Tuple, Union, Sequence

import numpy as np
import pandas as pd
from statsmodels.graphics.tsaplots import plot_acf as plot_acf_sm
from statsmodels.graphics.tsaplots import plot_pacf as plot_pacf_sm

from macrosynergy.visuals import FacetPlot


def plot_acf(
    df: pd.DataFrame,
    cids: List[str],
    xcat: str,
    lags: Union[int, Sequence] = 30,
    alpha: float = 0.05,
    remove_zero_predictor: bool = False,
    start: Optional[str] = None,
    end: Optional[str] = None,
    blacklist: Optional[Dict[str, List[str]]] = None,
    figsize: Tuple[float, float] = (16, 9),
    title: Optional[str] = None,
    share_x: bool = True,
    share_y: bool = True,
    zero: bool = False,
    auto_ylims: bool = True,
    **kwargs,
):
    """
    Plots a facet grid of autocorrelation functions for a given xcat and multiple cids.

    Parameters:
    -----------
    df : pd.DataFrame
        The input DataFrame with columns ['real_date', 'cid', 'xcat', 'value'].
    cids : List[str]
        List of cids to plot.
    xcat : str
        The xcat to filter and plot ACFs for.
    lags : Union[int, Sequence], default=30
        Number of lags for ACF calculation. If an integer, the lags from 1 to lags are plotted.
        If a sequence is provided, the lags are plotted as given.
    alpha : float, default=0.05
        Significance level for the confidence intervals.
    remove_zero_predictor : bool, default=False
        Remove zeros from the input series.
    blacklist : dict
        cross-sections with date ranges that should be excluded from the data frame. If
        one cross-section has several blacklist periods append numbers to the cross-section
        code.
    start : str
        ISO-8601 formatted date string. Select data from this date onwards. If None, all
        dates are selected.
    end : str
        ISO-8601 formatted date string. Select data up to and including this date. If
        None, all dates are selected.
    figsize : Tuple[float, float], default=(16,9)
        Figure size for the plot.
    title : Optional[str], default=None
        Title for the plot.
    share_x : bool, default=True
        Share x-axis across all subplots.
    share_y : bool, default=True
        Share y-axis across all subplots.
    zero : bool, default=False
        Include the zero lag in the plot.
    auto_ylims : bool, default=True
        Automatically set the y-axis limits for each subplot.
    kwargs : Dict
        Additional keyword arguments for the plot passed directly to Facetplot.lineplot.
    """

    _checks_plot_acf(
        df=df,
        cids=cids,
        xcat=xcat,
        lags=lags,
        alpha=alpha,
        remove_zero_predictor=remove_zero_predictor,
        start=start,
        end=end,
        blacklist=blacklist,
        figsize=figsize,
        title=title,
        share_x=share_x,
        share_y=share_y,
    )

    if title is None:
        title = f"Autocorrelation Function (ACF) for {xcat}"

    plot_func = _statsmodels_plot_acf_wrapper
    plot_func_kwargs = {
        "lags": lags,
        "alpha": alpha,
        "zero": zero,
        "auto_ylims": auto_ylims,
    }

    _plot_acf(
        df=df,
        cids=cids,
        xcat=xcat,
        plot_func=plot_func,
        plot_func_kwargs=plot_func_kwargs,
        remove_zero_predictor=remove_zero_predictor,
        start=start,
        end=end,
        blacklist=blacklist,
        figsize=figsize,
        title=title,
        share_x=share_x,
        share_y=share_y,
        **kwargs,
    )


def plot_pacf(
    df: pd.DataFrame,
    cids: List[str],
    xcat: str,
    lags: int = 30,
    alpha=0.05,
    remove_zero_predictor: bool = False,
    method="ywm",
    start: Optional[str] = None,
    end: Optional[str] = None,
    blacklist: Optional[Dict[str, List[str]]] = None,
    figsize: Tuple[float, float] = (16, 9),
    title: Optional[str] = None,
    share_x: bool = True,
    share_y: bool = True,
    zero: bool = False,
    auto_ylims: bool = True,
    **kwargs,
):
    """
    Plots a facet grid of partial autocorrelation functions for a given xcat and multiple cids.

    Parameters:
    -----------
    df : pd.DataFrame
        The input DataFrame with columns ['real_date', 'cid', 'xcat', 'value'].
    cids : List[str]
        List of cids to plot.
    xcat : str
        The xcat to filter and plot PACFs for.
    lags : Union[int, Sequence], default=30
        Number of lags for PACF calculation. If an integer, the lags from 1 to lags are plotted.
        If a sequence is provided, the lags are plotted as given.
    alpha : float, default=0.05
        Significance level for the confidence intervals.
    remove_zero_predictor : bool, default=False
        Remove zeros from the input series.
    method : str, default='ywm'
        Method for Statsmodel's PACF calculation. Must be one of ['ywm', 'ywmle', 'yw', 'ywadjusted', 'ols', 'ols-adjusted'].
    blacklist : dict
        cross-sections with date ranges that should be excluded from the data frame. If
        one cross-section has several blacklist periods append numbers to the cross-section
        code.
    start : str
        ISO-8601 formatted date string. Select data from this date onwards. If None, all
        dates are selected.
    end : str
        ISO-8601 formatted date string. Select data up to and including this date. If
        None, all dates are selected.
    figsize : Tuple[float, float], default=(16,9)
        Figure size for the plot.
    title : Optional[str], default=None
        Title for the plot.
    share_x : bool, default=True
        Share x-axis across all subplots.
    share_y : bool, default=True
        Share y-axis across all subplots.
    zero : bool, default=False
        Include the zero lag in the plot.
    auto_ylims : bool, default=True
        Automatically set the y-axis limits for each subplot.
    kwargs : Dict
        Additional keyword arguments for the plot passed directly to Facetplot.lineplot.
    """
    _checks_plot_acf(
        df=df,
        cids=cids,
        xcat=xcat,
        lags=lags,
        alpha=alpha,
        remove_zero_predictor=remove_zero_predictor,
        method=method,
        start=start,
        end=end,
        blacklist=blacklist,
        figsize=figsize,
        title=title,
        share_x=share_x,
        share_y=share_y,
    )

    if title is None:
        title = f"Partial Autocorrelation Function (PACF) for {xcat}"

    plot_func = _statsmodels_plot_pacf_wrapper
    plot_func_kwargs = {
        "lags": lags,
        "alpha": alpha,
        "method": method,
        "zero": zero,
        "auto_ylims": auto_ylims,
    }

    _plot_acf(
        df=df,
        cids=cids,
        xcat=xcat,
        plot_func=plot_func,
        plot_func_kwargs=plot_func_kwargs,
        remove_zero_predictor=remove_zero_predictor,
        start=start,
        end=end,
        blacklist=blacklist,
        figsize=figsize,
        title=title,
        share_x=share_x,
        share_y=share_y,
        **kwargs,
    )


def _plot_acf(
    df: pd.DataFrame,
    cids: List[str],
    xcat: str,
    plot_func: Callable,
    plot_func_kwargs: Dict,
    remove_zero_predictor: bool = False,
    start: Optional[str] = None,
    end: Optional[str] = None,
    blacklist: Optional[Dict[str, List[str]]] = None,
    figsize: Tuple[float, float] = (16, 9),
    title: Optional[str] = None,
    share_x: bool = True,
    share_y: bool = True,
    **kwargs,
):

    with FacetPlot(
        df=df,
        xcats=[xcat],
        cids=cids,
        intersect=True,
        start=start,
        end=end,
        blacklist=blacklist,
        tickers=None,
        metrics=["value"],
    ) as fp:

        if remove_zero_predictor:
            fp.df = fp.df.loc[fp.df["value"] != 0]

        if len(fp.cids) <= 3:
            kwargs["ncols"] = len(fp.cids)

        fp.cids = sorted(fp.cids)

        fp.lineplot(
            plot_func=plot_func,
            plot_func_kwargs=plot_func_kwargs,
            share_x=share_x,
            share_y=share_y,
            figsize=figsize,
            title=title,
            cid_grid=True,
            interpolate=True,
            **kwargs,
        )


def _statsmodels_plot_acf_wrapper(df, plt_dict, ax, **kwargs):
    """
    Wrapper function for statsmodels plot_acf.
    """
    y = plt_dict["Y"][0]
    cid, xcat = str(y).split("_", 1)
    selected_df = df.loc[cid, xcat]
    plot_acf_sm(x=selected_df["value"], ax=ax, title=cid, **kwargs)


def _statsmodels_plot_pacf_wrapper(df, plt_dict, ax, **kwargs):
    """
    Wrapper function for statsmodels plot_pacf.
    """
    y = plt_dict["Y"][0]
    cid, xcat = str(y).split("_", 1)
    selected_df = df.loc[cid, xcat]
    plot_pacf_sm(x=selected_df["value"], ax=ax, title=cid, **kwargs)


def _checks_plot_acf(
    df: pd.DataFrame,
    cids: List[str],
    xcat: str,
    lags: int = 30,
    alpha=0.05,
    remove_zero_predictor: bool = False,
    method="ywm",
    start: Optional[str] = None,
    end: Optional[str] = None,
    blacklist: Optional[Dict[str, List[str]]] = None,
    figsize: Tuple[float, float] = (16, 9),
    title: Optional[str] = None,
    share_x: bool = True,
    share_y: bool = True,
):
    if not isinstance(df, pd.DataFrame):
        raise TypeError("`df` must be a pandas DataFrame.")

    if len(df.columns) < 4:
        df = df.copy().reset_index()

    if not isinstance(lags, (int, np.ndarray, list, tuple)):
        raise TypeError("`lags` must be an integer.")

    if not isinstance(alpha, float):
        raise TypeError("`alpha` must be a number.")

    if not isinstance(remove_zero_predictor, bool):
        raise TypeError("`remove_zero_predictor` must be a boolean.")

    if start is None:
        start: str = pd.Timestamp(df["real_date"].min()).strftime("%Y-%m-%d")

    if end is None:
        end: str = pd.Timestamp(df["real_date"].max()).strftime("%Y-%m-%d")

    if not isinstance(xcat, str):
        raise TypeError("`xcat` must be a string.")

    if xcat not in df["xcat"].unique():
        raise ValueError(f"`xcat` {xcat} not found in the DataFrame.")

    if isinstance(cids, str):
        cids: List[str] = [cids]

    if not isinstance(cids, list):
        raise TypeError("`cids` must be a list.")

    if not all(isinstance(cid, str) for cid in cids):
        raise TypeError("All elements in `cids` must be strings.")

    if blacklist:
        if not isinstance(blacklist, dict):
            raise TypeError("`blacklist` must be a dictionary.")
        for key, value in blacklist.items():
            if not isinstance(key, str):
                raise TypeError("Keys in `blacklist` must be strings.")
            if not isinstance(value, list):
                raise TypeError("Values in `blacklist` must be lists.")

    valid_methods = ["ywm", "ywmle", "yw", "ywadjusted", "ols", "ols-adjusted"]
    if method not in valid_methods:
        raise ValueError(f"Invalid value for method. Must be one of {valid_methods}.")

    if not isinstance(figsize, tuple):
        raise TypeError("`figsize` must be a tuple.")

    if title is not None and not isinstance(title, str):
        raise TypeError("`title` must be a string.")

    if not isinstance(share_x, bool):
        raise TypeError("`share_x` must be a boolean.")

    if not isinstance(share_y, bool):
        raise TypeError("`share_y` must be a boolean.")


if __name__ == "__main__":
    import numpy as np

    from macrosynergy.management.simulate import make_test_df
    from macrosynergy.visuals import FacetPlot

    np.random.seed(42)

    cids: List[str] = [
        "USD",
        "EUR",
        "GBP",
        "AUD",
        "CAD",
        "JPY",
        "CHF",
        "NZD",
        "SEK",
        "NOK",
        "DKK",
        "INR",
    ]
    xcats: List[str] = [
        "FXXR",
        "EQXR",
        "RIR",
        "IR",
        "REER",
        "CPI",
        "PPI",
        "M2",
        "M1",
        "M0",
        "FXVOL",
        "FX",
    ]
    sel_cids: List[str] = [
        "USD",
        "EUR",
        "GBP",
        "AUD",
        "CAD",
        "JPY",
        "CHF",
        "NZD",
    ]  # ["USD", "EUR", "GBP"]
    sel_xcats: List[str] = ["FXXR", "EQXR", "RIR", "IR"]
    r_styles: List[str] = [
        "four-bit-sine",
        "decreasing-linear",
        "sharp-hill",
        "sine",
        "four-bit-sine",
    ]
    df: pd.DataFrame = make_test_df(
        cids=list(set(cids) - set(sel_cids)),
        xcats=xcats,
        start="2000-01-01",
    )

    for rstyle, xcatx in zip(r_styles, sel_xcats):
        dfB: pd.DataFrame = make_test_df(
            cids=sel_cids,
            xcats=[xcatx],
            start="2000-01-01",
            style=rstyle,
        )
        df: pd.DataFrame = pd.concat([df, dfB], axis=0)

    for ix, cidx in enumerate(sel_cids):
        df.loc[df["cid"] == cidx, "value"] = (
            ((df[df["cid"] == cidx]["value"]) * (ix + 1)).reset_index(drop=True).copy()
        )

    for ix, xcatx in enumerate(sel_xcats):
        df.loc[df["xcat"] == xcatx, "value"] = (
            ((df[df["xcat"] == xcatx]["value"]) * (ix * 10 + 1))
            .reset_index(drop=True)
            .copy()
        )

    df["value"] = df["value"] * (np.arange(len(df)) % 20 == 0)
    df["grading"] = np.nan

    plot_acf(
        df,
        cids=sel_cids,
        xcat="FXXR",
        # title="ACF Facet Plot",
        remove_zero_predictor=True,
        lags=[5, 6, 7],
        share_y=True,
    )

    plot_pacf(
        df,
        cids=sel_cids,
        xcat="FXXR",
        title="ACF Facet Plot",
        remove_zero_predictor=True,
        zero=True,
        lags=[5, 6, 7],
        share_y=True,
    )



"""
Class to produce point forecasts of returns given knowledge of an indicator state on a
specific date.
"""

import numpy as np
import pandas as pd
from sklearn.feature_selection import SelectorMixin
from sklearn.pipeline import Pipeline

from macrosynergy.learning.sequential import BasePanelLearner
from macrosynergy.management.utils import (
    concat_categorical,
    _insert_as_categorical,
    reduce_df,
)
from macrosynergy.management.types import QuantamentalDataFrame


class ReturnForecaster(BasePanelLearner):
    """
    Class to produce return forecasts for a single forward frequency, based on the
    indicator states at a specific date.

    Parameters
    ----------
    df : pd.DataFrame
        Daily quantamental dataframe in JPMaQS format containing a panel of features, as
        well as a panel of returns.
    xcats : list
        List comprising feature names, with the last element being the response variable
        name. The features and the response variable must be categories in the dataframe.
    real_date : str
        Date in ISO 8601 format at which time a forward forecast is made based on the
        information states on that day.
    cids : list, optional
        List of cross-section identifiers for consideration in the panel. Default is None,
        in which case all cross-sections in `df` are considered.
    blacklist : list, optional
        Blacklisting dictionary specifying date ranges for which cross-sectional
        information should be excluded. The keys are cross-sections and the values
        are tuples of start and end dates in ISO 8601 format. Default is None.
    freq : str, optional
        Frequency of the analysis. Default is "M" for monthly.
    lag : int, optional
        Number of periods to lag the response variable. Default is 1.
    xcat_aggs : list, optional
        List of aggregation functions to apply to the features, used when `freq` is not
        `D`. Default is ["last", "sum"].
    generate_labels : callable, optional
        Function to transform the response variable into either alternative regression
        targets or classification labels. Default is None.

    Notes
    -----
    This class is a simple interface to produce a single period forward
    forecast. The `real_date` parameter specifies the date of the information state used
    to generate the forecast. As an example, if the provided date is "2025-03-01", a
    monthly frequency is specified and the lag is 1, the information
    states on this date are set aside, and the previous data is downsampled to monthly
    (with the features lagged by 1 period).  On this dataset, model selection and fitting
    happen - and the forecast is produced for the single out-of-sample period (March 2025).
    """

    def __init__(
        self,
        df,
        xcats,
        real_date,
        cids=None,
        blacklist=None,
        freq="M",
        lag=1,
        xcat_aggs=["last", "sum"],
        generate_labels=None,
    ):
        # Parent checks
        self._check_init(
            df=df,
            xcats=xcats,
            cids=cids,
            start=None,
            end=None,
            blacklist=blacklist,
            freq=freq,
            lag=lag,
            xcat_aggs=xcat_aggs,
            generate_labels=generate_labels,
            drop_nas=True,
        )
        # Additional checks to those carried in the parent class
        if not isinstance(real_date, str):
            raise TypeError("The real_date argument must be a string.")
        try:
            pd.to_datetime(real_date)
        except ValueError:
            raise ValueError("'real_date' must be in ISO 8601 format.")
        
        self.real_date = pd.to_datetime(real_date)
        df_adj = reduce_df(df=df, xcats=xcats)
        self._check_factor_availability(df_adj, xcats, self.real_date)

        if not isinstance(lag, int):
            raise TypeError("The lag argument must be an integer.")
        if lag < 1:
            raise ValueError("The lag argument must be at least 1.")

        # Separate in-sample and out-of-sample data
        # NOTE: We include real_date in the training set due to the required
        # lag on the macro factors. In addition, the last available return will always
        # be that from the previous day. This means that no leakage is introduced.
        # This was included for the sake of a unit test.  
        df_train = df[df.real_date <= self.real_date]
        df_test = df[df.real_date == self.real_date]

        # Set up supervised learning training set
        super().__init__(
            df=df_train,
            xcats=xcats,
            cids=cids,
            start=None,
            end=None,
            blacklist=blacklist,
            freq=freq,
            lag=lag,
            xcat_aggs=xcat_aggs,
            generate_labels=generate_labels,
            drop_nas=True,
            skip_checks=True # So that the checks aren't run twice
        )

        # Set up out-of-sample dataset for forecasting
        self.X_test = (
            reduce_df(df=df_test, blacklist=blacklist)
            .pivot(index=["cid", "real_date"], columns="xcat", values="value")[
                self.X.columns
            ]
            .dropna()
            .sort_index()
        )
        self.unique_test_levels = self.X_test.index.get_level_values(0).unique()

        # Set up data structures for analytics
        self.preds = pd.DataFrame(columns=["real_date", "cid", "xcat", "value"]).astype(
            {
                "real_date": "datetime64[ns]",
                "cid": "category",
                "xcat": "category",
                "value": "float32",
            }
        )
        self.feature_importances = pd.DataFrame(
            columns=["real_date", "name"] + list(self.X.columns)
        ).astype(
            {
                **{col: "float32" for col in self.X.columns},
                "real_date": "datetime64[ns]",
                "name": "category",
            }
        )
        self.intercepts = pd.DataFrame(
            columns=["real_date", "name", "intercepts"]
        ).astype(
            {
                "real_date": "datetime64[ns]",
                "name": "category",
                "intercepts": "float32",
            }
        )
        self.selected_ftrs = pd.DataFrame(
            columns=["real_date", "name"] + list(self.X.columns)
        ).astype(
            {
                **{col: "int" for col in self.X.columns},
                "real_date": "datetime64[ns]",
                "name": "category",
            }
        )

        self.store_correlations = False
        # Create data structure to store correlation matrix of features feeding into the
        # final model and the input features themselves
        self.ftr_corr = pd.DataFrame(
            columns=[
                "real_date",
                "name",
                "predictor_input",
                "pipeline_input",
                "pearson",
            ]
        ).astype(
            {
                "real_date": "datetime64[ns]",
                "name": "category",
                "predictor_input": "category",
                "pipeline_input": "category",
                "pearson": "float",
            }
        )

    def calculate_predictions(
        self,
        name,
        models,
        hyperparameters,
        scorers,
        inner_splitters,
        search_type="grid",
        normalize_fold_results=False,
        cv_summary="mean",
        include_train_folds=False,
        n_iter=None,
        n_jobs_cv=1,
        n_jobs_model=1,
        store_correlations=False,
    ):
        """
        Calculate predictions for the out-of-sample period.

        Parameters
        ----------
        name : str
            Name of the signal optimization process.
        models : dict
            Dictionary of models to choose from. The keys are model names and the values
            are scikit-learn compatible models.
        hyperparameters : dict
            Dictionary of hyperparameters to choose from. The keys are model names and
            the values are hyperparameter dictionaries for the corresponding model. The
            keys must match with those provided in `models`.
        scorers : dict
            Dictionary of scoring functions to use in the hyperparameter optimization
            process. The keys are scorer names and the values are scikit-learn compatible
            scoring functions.
        inner_splitters : dict
            Dictionary of inner splitters to use in the hyperparameter optimization
            process. The keys are splitter names and the values are scikit-learn compatible
            cross-validator objects.
        search_type : str, optional
            Type of hyperparameter optimization to perform. Default is "grid". Options are
            "grid" and "prior".
        normalize_fold_results : bool, optional
            Whether to normalize the scores across folds before combining them. Default is
            False.
        cv_summary : str or callable, optional
            Summary function to use to combine scores across cross-validation folds.
            Default is "mean". Options are "mean", "median", "mean-std", "mean/std",
            "mean-std-ge" or a callable function.
        include_train_folds : bool, optional
            Whether to calculate cross-validation statistics on the training folds in 
            additional to the test folds. If True, the cross-validation estimator will be
            a function of both training data and test data. It is recommended to set 
            `cv_summary` appropriately. Default is False.
        n_iter : int, optional
            Number of iterations to run in random hyperparameter search. Default is None.
        n_jobs_cv : int, optional
            Number of parallel jobs to run the cross-validation process. Default is 1.
        n_jobs_model : int, optional
            Number of parallel jobs to run the model fitting process (if relevant).
            Default is 1.
        store_correlations : bool
            Whether to store the correlations between input pipeline features and input
            predictor features. Default is False.
        """
        # Checks
        self._check_run(
            name=name,
            outer_splitter=None,
            inner_splitters=inner_splitters,
            models=models,
            hyperparameters=hyperparameters,
            scorers=scorers,
            search_type=search_type,
            normalize_fold_results=normalize_fold_results,
            cv_summary=cv_summary,
            include_train_folds=include_train_folds,
            n_iter=n_iter,
            split_functions=None,
            n_jobs_outer=n_jobs_cv,
            n_jobs_inner=n_jobs_model,
        )
        if not isinstance(store_correlations, bool):
            raise TypeError("The store_correlations argument must be a boolean.")

        if store_correlations and not all(
            [isinstance(model, Pipeline) for model in models.values()]
        ):
            raise ValueError(
                "The store_correlations argument is only valid when all models are Scikit-learn Pipelines."
            )
        self.store_correlations = store_correlations

        # Get training and test indices
        train_idx = list(range(len(self.X)))
        base_splits = self._get_base_splits(inner_splitters)

        optim_results = self._worker(
            name=name,
            train_idx=train_idx,
            test_idx=[],
            inner_splitters=inner_splitters,
            models=models,
            hyperparameters=hyperparameters,
            scorers=scorers,
            cv_summary=cv_summary,
            include_train_folds=include_train_folds,
            search_type=search_type,
            normalize_fold_results=normalize_fold_results,
            n_iter=n_iter,
            n_jobs_inner=n_jobs_cv,
            base_splits=base_splits,
            n_splits_add=None,
            timestamp=self.real_date,
        )

        self._check_duplicate_results(name)

        # Collect results from the worker
        # quantamental_data, model_data, other_data
        model_choice_data = []
        ftr_coef_data = []
        intercept_data = []
        ftr_selection_data = []
        ftr_corr_data = []

        model_choice_data.append(optim_results["model_choice"])
        ftr_coef_data.append(optim_results["feature_importances"])
        intercept_data.append(optim_results["intercepts"])
        ftr_selection_data.append(optim_results["selected_ftrs"])
        ftr_corr_data.extend(optim_results["ftr_corr"])

        # First create pandas dataframes to store the forecasts
        forecasts_df = pd.DataFrame(
            index=self.X_test.index, columns=[name], data=np.nan, dtype="float32"
        )
        # Create quantamental dataframe of forecasts
        model = (
            models[optim_results["model_choice"][1]]
            .set_params(**optim_results["model_choice"][3])
            .fit(self.X, self.y)
        )
        forecasts = model.predict(self.X_test)
        forecasts_df.iloc[:, 0] = forecasts

        if self.blacklist is not None:
            for cross_section, periods in self.blacklist.items():
                cross_section_key = cross_section.split("_")[0]
                if cross_section_key in self.unique_test_levels:
                    forecasts_df.loc[
                        (cross_section_key, slice(periods[0], periods[1])), :
                    ] = np.nan

        forecasts_df.columns = forecasts_df.columns.astype("category")
        forecasts_df_long = pd.melt(
            frame=forecasts_df.reset_index(),
            id_vars=["real_date", "cid"],
            var_name="xcat",
        )
        self.preds = concat_categorical(
            df1=self.preds,
            df2=forecasts_df_long,
        )

        # Store model selection data
        model_df_long = pd.DataFrame(
            columns=[col for col in self.chosen_models.columns if col != "name"],
            data=model_choice_data,
        ).astype({"model_type": "category"})
        model_df_long = _insert_as_categorical(model_df_long, "name", name, 1)

        self.chosen_models = concat_categorical(
            df1=self.chosen_models,
            df2=model_df_long,
        )

        # Store feature coefficients
        coef_df_long = pd.DataFrame(
            columns=[col for col in self.feature_importances.columns if col != "name"],
            data=ftr_coef_data,
        )
        coef_df_long = _insert_as_categorical(coef_df_long, "name", name, 1)
        self.feature_importances = concat_categorical(
            self.feature_importances,
            coef_df_long,
        )

        # Store intercept
        intercept_df_long = pd.DataFrame(
            columns=[col for col in self.intercepts.columns if col != "name"],
            data=intercept_data,
        )
        intercept_df_long = _insert_as_categorical(intercept_df_long, "name", name, 1)
        self.intercepts = concat_categorical(
            self.intercepts,
            intercept_df_long,
        )

        # Store selected features
        ftr_select_df_long = pd.DataFrame(
            columns=[col for col in self.selected_ftrs.columns if col != "name"],
            data=ftr_selection_data,
        )
        ftr_select_df_long = _insert_as_categorical(ftr_select_df_long, "name", name, 1)
        self.selected_ftrs = concat_categorical(
            self.selected_ftrs,
            ftr_select_df_long,
        )

        ftr_corr_df_long = pd.DataFrame(
            columns=self.ftr_corr.columns, data=ftr_corr_data
        )

        self.ftr_corr = concat_categorical(
            self.ftr_corr,
            ftr_corr_df_long,
        )

    def _check_duplicate_results(self, name):
        conditions = [
            ("preds", "xcat", name),
            ("feature_importances", "name", name),
            ("intercepts", "name", name),
            ("selected_ftrs", "name", name),
            ("ftr_corr", "name", name),
            ("chosen_models", "name", name),
        ]
        self._remove_results(conditions)

    def store_split_data(
        self,
        pipeline_name,
        optimal_model,
        optimal_model_name,
        optimal_model_score,
        optimal_model_params,
        inner_splitters_adj,
        X_train,
        y_train,
        X_test,
        y_test,
        timestamp,
        adjusted_test_index,
    ):
        """
        Stores characteristics of the optimal model at each retraining date.

        Parameters
        ----------
        pipeline_name : str
            Name of the signal optimization process.
        optimal_model : RegressorMixin, ClassifierMixin or Pipeline
            Optimal model selected at each retraining date.
        optimal_model_name : str
            Name of the optimal model.
        optimal_model_score : float
            Cross-validation score for the optimal model.
        optimal_model_params : dict
            Chosen hyperparameters for the optimal model.
        inner_splitters_adj : dict
            Dictionary of adjusted inner splitters.
        X_train : pd.DataFrame
            Training feature matrix.
        y_train : pd.Series
            Training response variable.
        X_test : pd.DataFrame
            Test feature matrix.
        y_test : pd.Series
            Test response variable.
        timestamp : pd.Timestamp
            Timestamp of the retraining date.
        adjusted_test_index : pd.MultiIndex
            Adjusted test index to account for lagged features.

        Returns
        -------
        dict
            Dictionary containing feature importance scores, intercepts, selected features
            and correlations between inputs to pipelines and those entered into a final
            model.
        """
        feature_names = np.array(X_train.columns)
        if isinstance(optimal_model, Pipeline):
            final_estimator = optimal_model[-1]
            for _, transformer in reversed(optimal_model.steps):
                if isinstance(transformer, SelectorMixin):
                    feature_names = transformer.get_feature_names_out()
                    break
        else:
            final_estimator = optimal_model

        coefs = np.full(X_train.shape[1], np.nan)

        if hasattr(final_estimator, "feature_importances_") or (
            hasattr(final_estimator, "coef_")
        ):
            if hasattr(final_estimator, "feature_importances_"):
                coef = final_estimator.feature_importances_
            elif hasattr(final_estimator, "coef_"):
                coef = final_estimator.coef_
            # Reshape coefficients for storage compatibility
            if coef.ndim == 1:
                coefs = coef
            elif coef.ndim == 2:
                if coef.shape[0] == 1:
                    coefs = coef.flatten()

        coef_ftr_map = {ftr: coef for ftr, coef in zip(feature_names, coefs)}
        coefs = [
            coef_ftr_map[ftr] if ftr in coef_ftr_map else np.nan
            for ftr in X_train.columns
        ]
        if hasattr(final_estimator, "intercept_"):
            if isinstance(final_estimator.intercept_, np.ndarray):
                # Store the intercept if it has length one
                if len(final_estimator.intercept_) == 1:
                    intercepts = final_estimator.intercept_[0]
                else:
                    intercepts = np.nan
            else:
                # The intercept will be a float/integer
                intercepts = final_estimator.intercept_
        else:
            intercepts = np.nan

        # Get feature selection information
        if len(feature_names) == X_train.shape[1]:
            # Then all features were selected
            ftr_selection_data = [timestamp] + [1 for _ in feature_names]
        else:
            # Then some features were excluded
            ftr_selection_data = [timestamp] + [
                1 if name in feature_names else 0 for name in np.array(X_train.columns)
            ]

        ftr_corr_data = self._get_ftr_corr_data(
            pipeline_name, optimal_model, X_train, timestamp
        )

        # Store data
        split_result = {
            "feature_importances": [timestamp] + coefs,
            "intercepts": [timestamp, intercepts],
            "selected_ftrs": ftr_selection_data,
            "ftr_corr": ftr_corr_data,
        }

        return split_result

    def _get_ftr_corr_data(self, pipeline_name, optimal_model, X_train, timestamp):
        """
        Returns a list of correlations between the input features to a pipeline and the
        features inputted into the final model, at each retraining date.

        Parameters
        ----------
        pipeline_name : str
            Name of the signal optimization process.
        optimal_model : RegressorMixin, ClassifierMixin or Pipeline
            Optimal model selected at each retraining date.
        X_train : pd.DataFrame
            Input feature matrix.
        timestamp : pd.Timestamp
            Timestamp of the retraining date.

        Returns
        -------
        list
            List of correlations between the input features to a pipeline and the
            features inputted into the final model, at each retraining date.
        """
        if self.store_correlations and optimal_model is not None:
            # Transform the training data to the final feature space
            transformers = Pipeline(steps=optimal_model.steps[:-1])
            X_train_transformed = transformers.transform(X_train)
            n_features = X_train_transformed.shape[1]
            feature_names = (
                X_train_transformed.columns
                if isinstance(X_train_transformed, pd.DataFrame)
                else [f"Feature {i+1}" for i in range(n_features)]
            )
            # Calculate correlation between each original feature in X_train and
            # the transformed features in X_train_transformed
            if isinstance(X_train_transformed, pd.DataFrame):
                X_train_transformed = X_train_transformed.values

            ftr_corr_data = [
                [
                    timestamp,
                    pipeline_name,
                    final_feature_name,
                    input_feature_name,
                    np.corrcoef(
                        X_train_transformed[:, idx],
                        X_train[input_feature_name],
                    )[0, 1],
                ]
                for idx, final_feature_name in enumerate(feature_names)
                for input_feature_name in X_train.columns
            ]

        elif self.store_correlations and optimal_model is None:
            ftr_corr_data = [
                [
                    timestamp,
                    pipeline_name,
                    feature_name,
                    feature_name,
                    1,
                ]
                for feature_name in X_train.columns
            ]
        else:
            ftr_corr_data = []

        return ftr_corr_data

    def get_optimized_signals(self, name=None):
        """
        Returns forward forecasts for one or more pipelines.

        Parameters
        ----------
        name : str or list, optional
            Label(s) of forecast(s). Default is all stored in the
            class instance.

        Returns
        -------
        pd.DataFrame
            Pandas dataframe in JPMaQS format of working daily predictions.
        """
        if name is None:
            preds = self.preds
        else:
            if isinstance(name, str):
                name = [name]
            elif not isinstance(name, list):
                raise TypeError(
                    "The process name must be a string or a list of strings."
                )

            invalid_names = [n for n in name if n not in self.preds.xcat.unique()]
            if invalid_names:
                raise ValueError(
                    f"""The following process name(s) are not in the list of already-run
                    pipelines: {invalid_names}. Please check the names carefully. If
                    correct, please run calculate_predictions() first.
                    """
                )
            preds = self.preds[self.preds.xcat.isin(name)]

        signals_df = QuantamentalDataFrame(
            df=preds,
            categorical=self.df.InitializedAsCategorical,
        ).to_original_dtypes()

        return signals_df

    def get_selected_features(self, name=None):
        """
        Returns the selected features for one or more pipelines.

        Parameters
        ----------
        name: str or list, optional
            Label(s) of pipeline(s). Default is all stored in the
            class instance.

        Returns
        -------
        pd.DataFrame
            Pandas dataframe of the selected features at each retraining date.
        """
        if name is None:
            return self.selected_ftrs
        
        if isinstance(name, str):
            name = [name]
        elif not isinstance(name, list):
            raise TypeError(
                "The process name must be a string or a list of strings."
            )

        invalid_names = [n for n in name if n not in self.selected_ftrs.name.unique()]
        if invalid_names:
            raise ValueError(
                f"""The following process name(s) are not in the list of already-run
                pipelines: {invalid_names}. Please check the names carefully. If
                correct, please run calculate_predictions() first.
                """
            )
        return self.selected_ftrs[self.selected_ftrs.name.isin(name)]

    def get_feature_importances(self, name=None):
        """
        Returns feature importances for one or more pipelines.

        Parameters
        ----------
        name: str or list, optional
            Label(s) of pipeline(s). Default is all stored in the
            class instance.

        Returns
        -------
        pd.DataFrame
            Pandas dataframe of the feature importances, if available, learnt at each
            retraining date for a given pipeline.

        Notes
        -----
        Availability of feature importances is subject to the selected model having a
        `feature_importances_` or `coef_` attribute.
        """
        if name is None:
            return self.feature_importances

        if isinstance(name, str):
            name = [name]
        elif not isinstance(name, list):
            raise TypeError(
                "The process name must be a string or a list of strings."
            )

        invalid_names = [n for n in name if n not in self.feature_importances.name.unique()]
        if invalid_names:
            raise ValueError(
                f"""The following process name(s) are not in the list of already-run
                pipelines: {invalid_names}. Please check the names carefully. If
                correct, please run calculate_predictions() first.
                """
            )

        return self.feature_importances[
            self.feature_importances.name.isin(name)
        ].sort_values(by="real_date")

    def get_intercepts(self, name=None):
        """
        Returns intercepts for one or more pipelines.

        Parameters
        ----------
        name: str or list, optional
            Label(s) of pipeline(s). Default is all stored in the
            class instance.

        Returns
        -------
        pd.DataFrame
            Pandas dataframe of the intercepts, if available, learnt at each retraining
            date for a given pipeline.
        """
        if name is None:
            return self.intercepts

        if isinstance(name, str):
            name = [name]
        elif not isinstance(name, list):
            raise TypeError(
                "The process name must be a string or a list of strings."
            )
        
        invalid_names = [n for n in name if n not in self.intercepts.name.unique()]
        if invalid_names:
            raise ValueError(
                f"""The following process name(s) are not in the list of already-run
                pipelines: {invalid_names}. Please check the names carefully. If
                correct, please run calculate_predictions() first.
                """
            )

        return self.intercepts[self.intercepts.name.isin(name)].sort_values(
            by="real_date"
        )

    def get_feature_correlations(
        self,
        name=None,
    ):
        """
        Returns dataframe of feature correlations for one or more pipelines.

        Parameters
        ----------
        name: str or list, optional
            Label(s) of the pipeline(s). Default is all stored in the
            class instance.

        Returns
        -------
        pd.DataFrame
            Pandas dataframe of the correlations between the features passed into a model
            pipeline and the post-processed features inputted into the final model.
        """
        if name is None:
            return self.ftr_corr

        if isinstance(name, str):
            name = [name]
        elif not isinstance(name, list):
            raise TypeError(
                "The process name must be a string or a list of strings."
            )
        
        invalid_names = [n for n in name if n not in self.ftr_corr.name.unique()]
        if invalid_names:
            raise ValueError(
                f"""The following process name(s) are not in the list of already-run
                pipelines: {invalid_names}. Please check the names carefully. If
                correct, please run calculate_predictions() first.
                """
            )

        return self.ftr_corr[self.ftr_corr.name.isin(name)]

    def _check_factor_availability(self, df, xcats, real_date):
        """
        Check the date is in the dataframe and all categories are available on the date.
        """
        # Check that the date is in the span of the dataframe
        min_date = df.real_date.min()
        max_date = df.real_date.max()
        if real_date <= min_date or real_date > max_date:
            raise ValueError(
                f"Real date {real_date} is either not larger than the earliest date in the dataframe"
                " or nor smaller or equal to the latest date in the dataframe."
            )

        # Check that the date is in the dataframe
        if real_date not in df.real_date.unique():
            raise ValueError(f"Real date {real_date} is not in the dataframe.")

        # Check that all categories are available on the date
        num_categories = len(df[df.real_date == real_date].xcat.unique())
        if num_categories != len(xcats):
            raise ValueError(
                f"Not all categories are available on the real date {real_date}."
            )


if __name__ == "__main__":
    from macrosynergy.management.simulate import make_qdf
    from macrosynergy.management.types import QuantamentalDataFrame

    from macrosynergy.learning import (
        ExpandingKFoldPanelSplit,
        sharpe_ratio,
        MapSelector,
        PanelStandardScaler,
        PanelPCA,
    )

    from sklearn.metrics import make_scorer
    from sklearn.linear_model import Ridge, LinearRegression
    from sklearn.preprocessing import StandardScaler

    cids = ["AUD", "CAD", "GBP", "USD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]
    cols = ["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["CAD"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["GBP"] = ["2012-01-01", "2020-12-31", 0, 1]
    df_cids.loc["USD"] = ["2012-01-01", "2020-12-31", 0, 1]

    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc["XR"] = ["2012-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2012-01-01", "2020-12-31", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2012-01-01", "2020-12-31", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2012-01-01", "2020-12-31", -0.1, 2, 0.8, 0.3]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    # Initialize the return forecaster
    rf = ReturnForecaster(
        df=dfd,
        xcats=["CRY", "GROWTH", "INFL", "XR"],
        real_date="2020-12-31",
        freq="M",
        lag=1,
        xcat_aggs=["last", "sum"],
    )

    rf.calculate_predictions(
        name="ridge1",
        models={"Ridge": Pipeline([("scaler", StandardScaler()), ("ridge", Ridge())])},
        hyperparameters={"Ridge": {}},
        scorers={
            "sharpe": make_scorer(sharpe_ratio, greater_is_better=True),
        },
        inner_splitters={
            "Expanding": ExpandingKFoldPanelSplit(5),
        },
        cv_summary="mean-std",
    )

    rf.calculate_predictions(
        name="ridge100",
        models={
            "Ridge": Pipeline(
                [("scaler", StandardScaler()), ("ridge", Ridge(alpha=100))]
            )
        },
        hyperparameters={"Ridge": {}},
        scorers={
            "sharpe": make_scorer(sharpe_ratio, greater_is_better=True),
        },
        inner_splitters={
            "Expanding": ExpandingKFoldPanelSplit(5),
        },
        cv_summary="mean-std",
    )

    rf.calculate_predictions(
        name="var+lr",
        models={
            "Ridge": Pipeline(
                [
                    ("scaler", PanelStandardScaler()),
                    ("selector", MapSelector(n_factors=2)),
                    ("ridge", LinearRegression()),
                ]
            )
        },
        hyperparameters={"Ridge": {}},
        scorers={
            "sharpe": make_scorer(sharpe_ratio, greater_is_better=True),
        },
        inner_splitters={
            "Expanding": ExpandingKFoldPanelSplit(5),
        },
        cv_summary="mean-std",
    )

    rf.calculate_predictions(
        name="pca+lr",
        models={
            "Ridge": Pipeline(
                [
                    ("scaler", PanelStandardScaler()),
                    ("selector", PanelPCA(n_components=2, adjust_signs=True)),
                    ("ridge", LinearRegression()),
                ]
            )
        },
        hyperparameters={"Ridge": {}},
        scorers={
            "sharpe": make_scorer(sharpe_ratio, greater_is_better=True),
        },
        inner_splitters={
            "Expanding": ExpandingKFoldPanelSplit(5),
        },
        cv_summary="mean-std",
        store_correlations=True,
    )

    print(rf.get_optimized_signals())
    print(rf.get_feature_importances())
    print(rf.get_intercepts())
    print(rf.get_selected_features())
    print(rf.get_feature_correlations())
